#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This should probably be renamed guifront.py This defines all of the visual
components to the GUI It is invoked from guiback, which handles the nonvisual
logic.


BUGS:
    * Copying the ungrouped imageset raises an error. Should have the option
    to copy or move it. Other special imageset should not have this option.

    Should gray out an option if it is not available.


CommandLine:
    wbia --dbdir ~/lev/media/hdd/work/WWF_Lynx/ --name-tab
"""
import logging
from six.moves import zip, map, filter
from os.path import isdir
import sys
import functools
import utool as ut
from wbia.guitool.__PYQT__ import QtCore
from wbia.guitool.__PYQT__ import QtWidgets
from wbia.guitool.__PYQT__.QtCore import Qt
from wbia.guitool import slot_, ChangeLayoutContext

# from wbia.guitool import BlockContext
# from wbia.guitool import checks_qt_error
import wbia.guitool as gt
import wbia.plottool as pt
from wbia.plottool import color_funcs
from wbia.gui import guiheaders as gh
from wbia.gui import guimenus
import six
from wbia.viz.interact import interact_annotations2
from wbia.gui.guiheaders import (
    IMAGE_TABLE,
    IMAGE_GRID,
    ANNOTATION_TABLE,
    NAME_TABLE,
    NAMES_TREE,
    IMAGESET_TABLE,
)
from wbia.gui.models_and_views import (
    IBEISStripeModel,
    IBEISTableView,
    IBEISItemModel,
    IBEISTreeView,
    ImagesetTableModel,
    ImagesetTableView,
    IBEISTableWidget,
    IBEISTreeWidget,
    ImagesetTableWidget,
)
from wbia import constants as const

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


VERBOSE_GUI = ut.VERBOSE or ut.get_argflag(('--verbose-gui', '--verbgui'))
WITH_GUILOG = ut.get_argflag('--guilog')

# ############################
# ##### Tab Widgets #######
# ############################


class APITabWidget(QtWidgets.QTabWidget):
    """
    Holds the table-tabs

    use setCurrentIndex to change the selection
    """

    def __init__(tabwgt, parent=None, horizontalStretch=1):
        # QtWidgets.QTabWidget.__init__(tabwgt, parent)
        super(APITabWidget, tabwgt).__init__(parent)
        tabwgt.ibswgt = parent
        tabwgt._sizePolicy = gt.newSizePolicy(tabwgt, horizontalStretch=horizontalStretch)
        tabwgt.setSizePolicy(tabwgt._sizePolicy)
        # tabwgt.currentChanged.connect(tabwgt.setCurrentIndex)
        tabwgt.currentChanged.connect(tabwgt._on_tabletab_change)
        tabwgt.current_tblname = None

    @slot_(int)
    def _on_tabletab_change(tabwgt, index):
        """ Switch to the current imageset tab """
        if VERBOSE_GUI:
            logger.info('[apitab] _onchange(index=%r)' % (index,))
        tblname = tabwgt.ibswgt.tblname_list[index]
        tabwgt.current_tblname = tblname
        if VERBOSE_GUI:
            logger.info('[apitab] _onchange(tblname=%r)' % (tblname,))
        tabwgt.ibswgt.back._clear_selection()
        view = tabwgt.ibswgt.views[tblname]
        selected = view.selectionModel().selection()
        deselected = QtCore.QItemSelection()
        tabwgt.ibswgt.update_selection(selected, deselected)


class ImageSetTabWidget(QtWidgets.QTabWidget):
    """
    Handles tabs that contain individual image sets.
    """

    def __init__(imageset_tabwgt, parent=None, horizontalStretch=1):
        QtWidgets.QTabWidget.__init__(imageset_tabwgt, parent)
        imageset_tabwgt.ibswgt = parent
        imageset_tabwgt.setTabsClosable(True)
        imageset_tabwgt.setMaximumSize(9999, gt.get_cplat_tab_height())
        imageset_tabwgt.tabbar = imageset_tabwgt.tabBar()
        imageset_tabwgt.tabbar.setMovable(False)
        imageset_tabwgt.setStyleSheet('border: none;')
        imageset_tabwgt.tabbar.setStyleSheet('border: none;')
        sizePolicy = gt.newSizePolicy(
            imageset_tabwgt, horizontalStretch=horizontalStretch
        )
        imageset_tabwgt.setSizePolicy(sizePolicy)

        imageset_tabwgt.tabCloseRequested.connect(imageset_tabwgt._close_tab)
        imageset_tabwgt.currentChanged.connect(imageset_tabwgt._on_imagesettab_change)

        imageset_tabwgt.imgsetid_list = []
        # TURNING ON / OFF ALL IMAGES
        # imageset_tabwgt._add_imageset_tab(-1, const.ALL_IMAGE_IMAGESETTEXT)

    @slot_(int)
    def _on_imagesettab_change(imageset_tabwgt, index):
        """ Switch to the current imageset tab """
        if VERBOSE_GUI:
            logger.info('[imageset_tab_widget] _onchange(index=%r)' % (index,))
        if 0 <= index and index < len(imageset_tabwgt.imgsetid_list):
            imgsetid = imageset_tabwgt.imgsetid_list[index]
            if VERBOSE_GUI:
                logger.info('[IMAGESETTAB.ONCHANGE] imgsetid = %r' % (imgsetid,))
            imageset_tabwgt.ibswgt._change_imageset(imgsetid)
        else:
            imageset_tabwgt.ibswgt._change_imageset(-1)

    @slot_(int)
    def _close_tab(imageset_tabwgt, index):
        logger.info('[imageset_tab_widget] _close_tab(index=%r)' % (index,))
        if imageset_tabwgt.imgsetid_list[index] is not None:
            imageset_tabwgt.imgsetid_list.pop(index)
            imageset_tabwgt.removeTab(index)

    @slot_()
    def _close_all_tabs(imageset_tabwgt):
        if VERBOSE_GUI:
            logger.info('[imageset_tab_widget] _close_all_tabs()')
        while len(imageset_tabwgt.imgsetid_list) > 0:
            index = 0
            imageset_tabwgt.imgsetid_list.pop(index)
            imageset_tabwgt.removeTab(index)

    @slot_(int)
    def _close_tab_with_imgsetid(imageset_tabwgt, imgsetid):
        logger.info(
            '[imageset_tab_widget] _close_tab_with_imgsetid(imgsetid=%r)' % (imgsetid)
        )
        try:
            index = imageset_tabwgt.imgsetid_list.index(imgsetid)
            imageset_tabwgt._close_tab(index)
        except Exception:
            pass

    def _add_imageset_tab(imageset_tabwgt, imgsetid, imagesettext):
        if VERBOSE_GUI:
            logger.info(
                '[_add_imageset_tab] imgsetid=%r, imagesettext=%r'
                % (imgsetid, imagesettext)
            )
        if imgsetid not in imageset_tabwgt.imgsetid_list:
            if VERBOSE_GUI:
                logger.info('[_add_imageset_tab] adding new image tab')
            imageset_tabwgt.imgsetid_list.append(imgsetid)
            index = len(imageset_tabwgt.imgsetid_list) - 1
            tab_name = str(imagesettext)
            # Only has a tab, doesn't actually contain anything
            hack_newtab = QtWidgets.QWidget(parent=imageset_tabwgt)
            imageset_tabwgt.addTab(hack_newtab, tab_name)
        else:
            if VERBOSE_GUI:
                logger.info('[_add_imageset_tab] using existing image tab')
            index = imageset_tabwgt.imgsetid_list.index(imgsetid)

        if VERBOSE_GUI:
            logger.info('[_add_imageset_tab] setCurrentIndex(index=%r)' % (index,))
        imageset_tabwgt.setCurrentIndex(index)
        # Dont call this, it is triggered twice
        # imageset_tabwgt._on_imagesettab_change(index)

    def _update_imageset_tab_name(imageset_tabwgt, imgsetid, imagesettext):
        for index, _id in enumerate(imageset_tabwgt.imgsetid_list):
            if imgsetid == _id:
                imageset_tabwgt.setTabText(index, imagesettext)


# ############################
# ####### Main Widget ########
# ############################


class IBEISMainWindow(QtWidgets.QMainWindow):
    quitSignal = QtCore.pyqtSignal()
    dropSignal = QtCore.pyqtSignal(list)

    def __init__(mainwin, back=None, ibs=None, parent=None):
        super(IBEISMainWindow, mainwin).__init__(parent)
        # Menus
        try:
            mainwin.setUnifiedTitleAndToolBarOnMac(False)
        except AttributeError as ex:
            ut.printex(
                ex, 'setUnifiedTitleAndToolBarOnMac is not working', iswarning=True
            )
        guimenus.setup_menus(mainwin, back)
        # Central Widget
        mainwin.ibswgt = IBEISGuiWidget(back=back, ibs=ibs, parent=mainwin)
        mainwin.setCentralWidget(mainwin.ibswgt)
        mainwin.setAcceptDrops(True)
        if back is not None:
            mainwin.quitSignal.connect(back.quit)
        else:
            raise AssertionError('need backend')
        mainwin.dropSignal.connect(mainwin.ibswgt.imagesDropped)
        #
        mainwin.resize(900, 600)

    @slot_()
    def closeEvent(mainwin, event):
        event.accept()
        mainwin.quitSignal.emit()

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            links = []
            for url in event.mimeData().urls():
                links.append(str(url.toLocalFile()))
            self.dropSignal.emit(links)
        else:
            event.ignore()

    @slot_()
    def expand_names_tree(mainwin):
        view = mainwin.ibswgt.views[gh.NAMES_TREE]
        view.expandAll()


# ############################
# #### IBEIS GUI Widget ######
# ############################


IBEIS_WIDGET_BASE = QtWidgets.QWidget


class IBEISGuiWidget(IBEIS_WIDGET_BASE):
    """
    CommandLine:
        # Testing
        python -m wbia --db NNP_Master3 --onlyimgtbl
        python -m wbia --db PZ_Master1 --onlyimgtbl
    """

    def __init__(ibswgt, back=None, ibs=None, parent=None):
        super(IBEISGuiWidget, ibswgt).__init__(parent)
        ibswgt.ibs = ibs
        ibswgt.back = back
        # Structures that will hold models and views
        ibswgt.models = {}
        ibswgt.views = {}

        # FIXME: Duplicate models
        # Create models and views
        # Define the abstract item models and views for the tables
        ibswgt.tblname_list = []
        ibswgt.modelview_defs = []

        # NEW DYNAMIC WAY OF USING TABS AND API VIEWS
        # ADD IMAGE TABLE
        if True:
            ibswgt.tblname_list.append(IMAGE_TABLE)
            ibswgt.modelview_defs.append(
                (IMAGE_TABLE, IBEISTableWidget, IBEISItemModel, IBEISTableView)
            )
        # ADD IMAGE GRID
        USE_GRID = False
        if USE_GRID and not ut.get_argflag('--onlyimgtbl'):
            ibswgt.tblname_list.append(IMAGE_GRID)
            ibswgt.modelview_defs.append(
                (IMAGE_GRID, IBEISTableWidget, IBEISStripeModel, IBEISTableView)
            )
        # ADD ANNOT GRID
        if not (ut.get_argflag('--noannottbl') or ut.get_argflag('--onlyimgtbl')):
            ibswgt.tblname_list.append(gh.ANNOTATION_TABLE)
            ibswgt.modelview_defs.append(
                (gh.ANNOTATION_TABLE, IBEISTableWidget, IBEISItemModel, IBEISTableView)
            )
        # ADD NAME TREE
        if not (ut.get_argflag('--nonametree') or ut.get_argflag('--onlyimgtbl')):
            ibswgt.tblname_list.append(NAMES_TREE)
            ibswgt.modelview_defs.append(
                (NAMES_TREE, IBEISTreeWidget, IBEISItemModel, IBEISTreeView)
            )
        # ADD IMAGESET TABLE
        ibswgt.super_tblname_list = ibswgt.tblname_list + [IMAGESET_TABLE]
        ibswgt.modelview_defs.append(
            (IMAGESET_TABLE, ImagesetTableWidget, ImagesetTableModel, ImagesetTableView)
        )

        # DO INITALIZATION
        # Create and layout components
        ibswgt._init_components()
        ibswgt._connect_signals_and_slots()
        ibswgt.connect_wbia_control(ibswgt.ibs)

        # Lazy load data every so often
        ibswgt.data_load_timer = QtCore.QTimer()
        ibswgt.data_load_timer.timeout.connect(ibswgt.data_load_loop)
        ibswgt.data_load_freq = 1000
        # ibswgt.tt = None
        ibswgt.data_load_timer.start(ibswgt.data_load_freq)

    # @QtCore.pyqtSlot
    def data_load_loop(ibswgt):
        # Get the current view
        view = ibswgt._tables_tab_widget.currentWidget()
        model = view.model()
        if model.canFetchMore(None):
            model.fetchMore(None)
        # if ibswgt.tt is None:
        #    ellapsed = ibswgt.data_load_freq
        # else:
        #    ellapsed = ut.toc(ibswgt.tt)
        # logger.info('Load more data for %r' % (view,))
        # # freq = ibswgt.data_load_freq / (1000 * 2)
        # # frac = freq / ellapsed
        # # logger.info('ibswgt.data_load_freq = %r' % (freq,))
        # # logger.info('ellapsed = %r' % (ellapsed,))
        # # logger.info('frac = %r' % (frac,))
        # # if model.batch_size is not None:
        #    #new_batch_size = model.batch_size
        #    #new_batch_size = int(new_batch_size * frac * 2)
        #    #new_batch_size = max(new_batch_size * 2, 2)
        #    #new_batch_size = min(model.batch_size * 2, new_batch_size)
        #    #new_batch_size = min(100, new_batch_size)
        #    #model.batch_size = new_batch_size
        #    #logger.info('model.batch_size = %r' % (model.batch_size,))
        # # ibswgt.tt = ut.tic()
        # else:
        #    ibswgt.tt = None

    def _init_components(ibswgt):
        """ Defines gui components and inits layout """
        # Layout
        ibswgt.vlayout = QtWidgets.QVBoxLayout(ibswgt)
        ibswgt.setLayout(ibswgt.vlayout)

        ibswgt.vsplitter = gt.newSplitter(ibswgt, orientation=Qt.Vertical)
        ibswgt.hsplitter = gt.newSplitter(
            ibswgt, orientation=Qt.Horizontal, verticalStretch=18
        )
        #
        # Tables Tab
        ibswgt._tables_tab_widget = APITabWidget(parent=ibswgt, horizontalStretch=81)
        for tblname, WidgetClass, ModelClass, ViewClass in ibswgt.modelview_defs:
            # Make view first to pass as parent
            ibswgt.views[tblname] = ViewClass(parent=ibswgt)
            # FIXME: It is very bad to give the model a view. Only the view
            # should have a model
            ibswgt.models[tblname] = ModelClass(parent=ibswgt.views[tblname])
        # Connect models and views
        for tblname in ibswgt.super_tblname_list:
            ibswgt.views[tblname].setModel(ibswgt.models[tblname])
        # Add Image, ANNOTATION, and Names as tabs
        for tblname in ibswgt.tblname_list:
            ibswgt._tables_tab_widget.addTab(ibswgt.views[tblname], tblname)

        # Force full loading
        ibswgt.models[IMAGE_TABLE].batch_size = 1000
        ibswgt.models[IMAGESET_TABLE].batch_size = 1000
        ibswgt.models[ANNOTATION_TABLE].batch_size = 1000
        ibswgt.models[NAMES_TREE].batch_size = 2
        ibswgt.models[NAMES_TREE].batch_size = 100

        # Custom ImageSet Tab Wiget
        ibswgt.imageset_tabwgt = ImageSetTabWidget(parent=ibswgt)
        ibswgt.vlayout.addWidget(ibswgt.vsplitter)
        ibswgt.vsplitter.addWidget(ibswgt.hsplitter)

        ibswgt.status_wgt = status_wgt = gt.newWidget(
            ibswgt.vsplitter,
            orientation=Qt.Vertical,
            spacing=3,
            margin=0,
            name='StatusWidget',
        )
        ibswgt.vsplitter.addWidget(status_wgt)

        # On the LEFT add the the table of ImageSets
        imgset_table_view = ibswgt.views[IMAGESET_TABLE]
        imgset_table_view.setSizePolicy(
            gt.newSizePolicy(hSizePolicy='Expanding', hStretch=2)
        )

        # On the RIGHT add the DataTables and tabs
        right_hack_wgt = gt.newWidget()
        right_hack_wgt.addWidget(ibswgt.imageset_tabwgt)
        right_hack_wgt.addWidget(ibswgt._tables_tab_widget)
        right_hack_wgt.setSizePolicy(gt.newSizePolicy(horizontalStretch=5))

        # Hack because the tables aren't actually belonging to the tabs
        # They are just controlled by the change
        ibswgt.hsplitter.addWidget(imgset_table_view)
        ibswgt.hsplitter.addWidget(right_hack_wgt)

        _NEWLBL = functools.partial(gt.newLabel, ibswgt)
        _NEWBUT = functools.partial(gt.newButton, ibswgt)
        # _COMBO  = functools.partial(gt.newComboBox, ibswgt)
        _NEWTEXT = functools.partial(gt.newLineEdit, ibswgt, verticalStretch=1)

        primary_fontkw = dict(bold=True, pointSize=11)
        secondary_fontkw = dict(bold=False, pointSize=9)
        # advanced_fontkw = dict(bold=False, pointSize=8, italic=True)
        identify_color = (255, 150, 0)

        ibswgt.tablename_to_status_widget_index = {
            IMAGESET_TABLE: 1,
            IMAGE_TABLE: 3,
            IMAGE_GRID: 3,
            gh.ANNOTATION_TABLE: 5,
            NAMES_TREE: 7,
            NAME_TABLE: 7,
        }
        ibswgt.key_to_objnice = {
            IMAGESET_TABLE: 'ImageSet',
            IMAGE_TABLE: 'Image',
            ANNOTATION_TABLE: 'Annotation',
            NAMES_TREE: 'Name',
        }
        ibswgt.status_widget_list = [
            _NEWLBL('Selected ImageSet: ', fontkw=secondary_fontkw, align='right'),
            _NEWTEXT(enabled=True, readOnly=True),
            ##
            _NEWLBL('Selected Image: ', fontkw=secondary_fontkw, align='right'),
            _NEWTEXT(
                enabled=True,
                readOnly=False,
                editingFinishedSlot=ibswgt.select_image_text_editing_finished,
            ),
            ##
            _NEWLBL('Selected Annotation: ', fontkw=secondary_fontkw, align='right'),
            _NEWTEXT(
                enabled=True,
                readOnly=False,
                editingFinishedSlot=ibswgt.select_annot_text_editing_finished,
            ),
            ##
            _NEWLBL('Selected Name: ', fontkw=secondary_fontkw, align='right'),
            _NEWTEXT(
                enabled=True,
                readOnly=False,
                editingFinishedSlot=ibswgt.select_name_text_editing_finished,
            ),
        ]

        back = ibswgt.back

        ibswgt.batch_intra_occurrence_query_button = _NEWBUT(
            '4) ID Encounters',
            functools.partial(
                back.compute_queries,
                daids_mode=const.INTRA_OCCUR_KEY,
                query_is_known=None,
                use_prioritized_name_subset=False,
                cfgdict={'can_match_samename': False, 'use_k_padding': False},
            ),
            bgcolor=color_funcs.adjust_hsv_of_rgb255(identify_color, -0.01, -0.7, 0.0),
            fgcolor=(0, 0, 0),
            fontkw=primary_fontkw,
        )

        ibswgt.batch_vsexemplar_query_button = _NEWBUT(
            '5) ID Exemplars',
            functools.partial(
                back.compute_queries,
                daids_mode=const.VS_EXEMPLARS_KEY,
                use_prioritized_name_subset=True,
                query_is_known=None,
                cfgdict={'can_match_samename': False, 'use_k_padding': False},
            ),
            bgcolor=color_funcs.adjust_hsv_of_rgb255(identify_color, -0.02, -0.7, 0.0),
            fgcolor=(0, 0, 0),
            fontkw=primary_fontkw,
        )

        ibswgt.import_button = _NEWBUT(
            '1) Import',
            # back.import_images_from_dir,
            back.import_button_click,
            bgcolor=(235, 200, 200),
            fontkw=primary_fontkw,
        )

        ibswgt.imageset_button = _NEWBUT(
            '2) Group',
            ibswgt.back.do_group_occurrence_step,
            bgcolor=(255, 255, 150),
            fontkw=primary_fontkw,
        )

        ibswgt.detect_button = _NEWBUT(
            '3) Detect',
            ibswgt.back.run_detection_step,
            bgcolor=(150, 255, 150),
            fontkw=primary_fontkw,
        )

        ibswgt.reviewed_button = _NEWBUT(
            '6) Complete',
            ibswgt.back.commit_to_wb_step,
            bgcolor=color_funcs.adjust_hsv_of_rgb255((0, 232, 211), 0.0, -0.9, 0.0),
            fontkw=primary_fontkw,
            enabled=True,
        )

        ibswgt.control_widget_lists = [
            [
                ibswgt.import_button,
                ibswgt.imageset_button,
                _NEWLBL(''),
                ibswgt.detect_button,
                ibswgt.batch_intra_occurrence_query_button,
                ibswgt.batch_vsexemplar_query_button,
                ibswgt.reviewed_button,
            ],
            [
                _NEWBUT(
                    'Advanced ID Interface',
                    back.show_advanced_id_interface,
                    bgcolor=color_funcs.adjust_hsv_of_rgb255(identify_color),
                    fgcolor=(0, 0, 0),
                    fontkw=primary_fontkw,
                )
            ],
        ]

        # Other components
        # New widget has black magic (for implicit layouts) in it
        # Add control widgets (import, group, species selector, etc...)
        for count, control_widgets in enumerate(ibswgt.control_widget_lists):
            _container = status_wgt.addNewWidget(
                orientation=Qt.Horizontal, margin=0, name='ControlContainer%d' % (count,)
            )
            for widget in control_widgets:
                _container.addWidget(widget)

        # Output log (turned off by default)
        ibswgt.outputLog = gt.newOutputLog(
            status_wgt, pointSize=8, visible=WITH_GUILOG, verticalStretch=6
        )
        status_wgt.addWidget(ibswgt.outputLog)

        # Add selected ids status widget
        ibswgt.selectionStatusWidget = status_wgt.addNewWidget(
            orientation=Qt.Horizontal, margin=3, name='SelectionStatus'
        )
        for widget in ibswgt.status_widget_list:
            ibswgt.selectionStatusWidget.addWidget(widget)

        # Progress bar is at the bottom
        ibswgt.prog_bar = status_wgt.addNewProgressBar(
            visible=False, verticalStretch=1, name='prog_bar'
        )

        # ibswgt.vsplitter.print_widget_heirarchy(max_depth=4)

    def _connect_signals_and_slots(ibswgt):
        if VERBOSE_GUI:
            logger.info('[newgui] _connect_signals_and_slots')
        for tblname in ibswgt.super_tblname_list:
            tblview = ibswgt.views[tblname]
            tblview.doubleClicked.connect(ibswgt.on_doubleclick)
            tblview.contextMenuClicked.connect(ibswgt.on_contextMenuClicked)
            if tblname != gh.IMAGESET_TABLE:
                tblview.selectionModel().selectionChanged.connect(ibswgt.update_selection)
            # front.printSignal.connect(back.backend_print)
            # front.raiseExceptionSignal.connect(back.backend_exception)
            # CONNECT HOOK TO GET NUM ROWS
            tblview.rows_updated.connect(ibswgt.on_rows_updated)

    @slot_(QtCore.QItemSelection, QtCore.QItemSelection)
    def update_selection(ibswgt, selected, deselected):
        """
        Quirky behavior: if you select two columns in a row and then unselect
        only one, the whole row is unselected, because this function only deals
        with deltas.

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.gui.newgui import *  # NOQA
            >>> ibs, back, ibswgt, testdata_main_loop = testdata_guifront()
            >>> ibswgt.set_table_tab(gh.NAMES_TREE)
            >>> view = ibswgt.views[gh.NAMES_TREE]
            >>> view.expandAll()
            >>> AUTOSELECT = False
            >>> if AUTOSELECT:
            ...     view.selectAll()
            >>> selmodel = view.selectionModel()
            >>> selected = selmodel.selection()
            >>> deselected = QtCore.QItemSelection()
            >>> # verify results
            >>> print(result)

        """
        # logger.info('[ibswgt] update selection')
        # logger.info('selected = ' + str(selected.indexes()))
        # logger.info('deselected = ' + str(deselected.indexes()))
        deselected_model_index_list_ = deselected.indexes()
        selected_model_index_list_ = selected.indexes()

        def get_selection_info(model_index_list_):
            model_index_list = [
                qtindex for qtindex in model_index_list_ if qtindex.isValid()
            ]
            model_list = [qtindex.model() for qtindex in model_index_list]
            tablename_list = [model.name for model in model_list]
            level_list = [
                model._get_level(qtindex)
                for model, qtindex in zip(model_list, model_index_list)
            ]
            rowid_list = [
                model._get_row_id(qtindex)
                for model, qtindex in zip(model_list, model_index_list)
            ]
            table_key_list = list(zip(tablename_list, level_list))
            return table_key_list, rowid_list

        select_table_key_list, select_rowid_list = get_selection_info(
            selected_model_index_list_
        )
        deselect_table_key_list, deselect_rowid_list = get_selection_info(
            deselected_model_index_list_
        )

        table_key2_selected_rowids = dict(
            ut.group_items(select_rowid_list, select_table_key_list)
        )
        table_key2_deselected_rowids = dict(
            ut.group_items(deselect_rowid_list, deselect_table_key_list)
        )

        table_key2_selected_rowids = {
            key: list(set(val)) for key, val in six.iteritems(table_key2_selected_rowids)
        }
        table_key2_deselected_rowids = {
            key: list(set(val))
            for key, val in six.iteritems(table_key2_deselected_rowids)
        }
        if ut.VERBOSE:
            logger.info(
                'table_key2_selected_rowids = ' + ut.repr2(table_key2_selected_rowids)
            )
            logger.info(
                'table_key2_deselected_rowids = ' + ut.repr2(table_key2_deselected_rowids)
            )

        gh_const_tablename_map = {
            (IMAGE_TABLE, 0): const.IMAGE_TABLE,
            (IMAGE_GRID, 0): const.IMAGE_TABLE,
            (gh.ANNOTATION_TABLE, 0): const.ANNOTATION_TABLE,
            (NAME_TABLE, 0): const.NAME_TABLE,
            (NAMES_TREE, 0): const.NAME_TABLE,
            (NAMES_TREE, 1): const.ANNOTATION_TABLE,
        }
        # here tablename is a backend const tablename
        for table_key, id_list in six.iteritems(table_key2_deselected_rowids):
            tablename = gh_const_tablename_map[table_key]
            ibswgt.back._set_selection3(tablename, id_list, mode='diff')
        for table_key, id_list in six.iteritems(table_key2_selected_rowids):
            tablename = gh_const_tablename_map[table_key]
            ibswgt.back._set_selection3(tablename, id_list, mode='add')
        ibswgt.back.update_selection_texts()

        # tblview.selectionModel().selectedIndexes()

    def changing_models_gen(ibswgt, tblnames=None):
        """
        Loops over tablenames emitting layoutChanged at the end for each
        """
        tblnames = ibswgt.super_tblname_list if tblnames is None else tblnames
        if VERBOSE_GUI:
            logger.info('[newgui] changing_models_gen(tblnames=%r)' % (tblnames,))
        model_list = [ibswgt.models[tblname] for tblname in tblnames]
        # model_list = [ibswgt.models[tblname] for tblname in tblnames if
        # ibswgt.views[tblname].isVisible()]
        with ChangeLayoutContext(model_list):
            for tblname in tblnames:
                yield tblname

    def update_tables(ibswgt, tblnames=None, clear_view_selection=True):
        """ forces changing models """
        logger.info('[newgui] update_tables(%r)' % (tblnames,))
        hack_selections = []
        # logger.info('[new_gui.UPDATE_TABLES]')
        for tblname in ibswgt.changing_models_gen(tblnames=tblnames):
            # logger.info('[new_gui.update_tables] tblname=%r' % (tblname, ))
            model = ibswgt.models[tblname]
            view = ibswgt.views[tblname]
            if clear_view_selection:
                hack_selections.append(view.clearSelection)
            model._update()
        # Hack: Call this outside changing models gen
        for clearSelection in hack_selections:
            clearSelection()

    def connect_wbia_control(ibswgt, ibs):
        """ Connects a new ibscontroler to the models """
        if VERBOSE_GUI:
            logger.info('[newgui] connect_wbia_control(ibs=%r)' % (ibs,))
        ibswgt.imageset_tabwgt._close_all_tabs()
        if ibs is None:
            if VERBOSE_GUI:
                logger.info('[newgui] invalid ibs')
            title = 'No Database Opened'
            ibswgt.setWindowTitle(title)
        else:
            if VERBOSE_GUI:
                logger.info('[newgui] Connecting valid ibs=%r' % ibs.get_dbname())
            # with ut.Indenter('[CONNECTING]'):
            # Give the frontend the new control
            ibswgt.ibs = ibs
            if not ut.get_argflag('--fast'):
                with ut.Timer('update special', verbose=VERBOSE_GUI):
                    if not ibs.readonly:
                        ibs.update_special_imagesets()
            else:
                if VERBOSE_GUI:
                    logger.info('Skipping special imagesets')
            # Update the api models to use the new control
            with ut.Timer('make headers', verbose=VERBOSE_GUI):
                header_dict, declare_tup = gh.make_wbia_headers_dict(ibswgt.ibs)
            ibswgt.declare_tup = declare_tup

            def get_title(ibs):
                """
                DEPRICATE OR MOVE
                """
                if ibs is None:
                    title = 'IBEIS - No Database Directory Open'
                elif ibs.dbdir is None:
                    title = 'IBEIS - !! INVALID DATABASE !!'
                else:
                    dbdir = ibs.get_dbdir()
                    dbname = ibs.get_dbname()
                    title = 'IBEIS - %r - Database Directory = %s' % (dbname, dbdir)
                    wb_target = ibs.const.WILDBOOK_TARGET
                    # params.args.wildbook_target
                    if wb_target is not None:
                        title = '%s - Wildbook Target = %s' % (title, wb_target)
                return title

            title = get_title(ibswgt.ibs)
            ibswgt.setWindowTitle(title)
            if ut.VERBOSE:
                logger.info('[newgui] Calling model _update_headers')
            # block_wgt_flag = ibswgt._tables_tab_widget.blockSignals(True)

            with ut.Timer('[newgui] update models', verbose=VERBOSE_GUI):
                # for tblname in ibswgt.changing_models_gen(ibswgt.super_tblname_list):
                for tblname in ibswgt.super_tblname_list:
                    model = ibswgt.models[tblname]
                    view = ibswgt.views[tblname]
                    # if not view.isVisible():
                    #    logger.info(view)
                    # ut.embed()
                    header = header_dict[tblname]
                    # widget = ibswgt.widgets[tblname]
                    # widget.change_headers(header)
                    # NOT SURE IF THESE BLOCKERS SHOULD BE COMMENTED
                    # block_model_flag = model.blockSignals(True)
                    model._update_headers(**header)
                    view._update_headers(**header)  # should use model headers
                    # model.blockSignals(block_model_flag)
                    #
                    # view.infer_delegates_from_model()
                for tblname in ibswgt.super_tblname_list:
                    view = ibswgt.views[tblname]
                    # if not view.isVisible():
                    #    logger.info(view)
                    #    continue
                    view.hide_cols()
            # ibswgt._tables_tab_widget.blockSignals(block_wgt_flag)

            # Update species with ones enabled in database
            if not ibs.readonly:
                ibswgt.update_species_available()

            # FIXME: bad code
            # TODO: load previously loaded imageset or nothing
            LOAD_IMAGESET_ON_START = not ut.get_argflag('--fast')
            if LOAD_IMAGESET_ON_START:
                imgsetid_list = ibs.get_valid_imgsetids(shipped=False)
                if len(imgsetid_list) > 0:
                    DEFAULT_LARGEST_IMAGESET = False
                    if DEFAULT_LARGEST_IMAGESET:
                        numImg_list = ibs.get_imageset_num_gids(imgsetid_list)
                        argx = ut.list_argsort(numImg_list, reverse=True)[0]
                        imgsetid = imgsetid_list[argx]
                    else:  # Grab "first" imageset
                        imgsetid = imgsetid_list[0]
                    # ibswgt._change_imageset(imgsetid)
                    ibswgt.select_imageset_tab(imgsetid)
                else:
                    ibswgt._change_imageset(-1)

    def update_species_available(
        ibswgt, reselect=False, reselect_new_name=None, deleting=False
    ):
        ibs = ibswgt.ibs
        # TODO: update these options depending on ibs.get_species_with_detectors
        # when a controller is attached to the gui
        detection_combo_box_options = [
            # Text              # Value
            # ('Select Species',  'none'),
            ('Select Species', const.UNKNOWN),
            ('Unknown', const.UNKNOWN),
            # 'none'),
        ] + sorted(list(ibs.get_working_species()))
        species_text = ibswgt.back.get_selected_species()
        reselect_index = None
        if not deleting and reselect_new_name is None and species_text is not None:
            species_rowid = ibs.get_species_rowids_from_text(species_text)
            reselect_new_name = ibs.get_species_nice(species_rowid)
            if VERBOSE_GUI:
                logger.info(
                    '[front] Reselecting old selection: %r' % (reselect_new_name,)
                )
        nice_name_list = [str(_[0]) for _ in detection_combo_box_options]
        if reselect_new_name in nice_name_list:
            reselect_index = nice_name_list.index(reselect_new_name)
            if VERBOSE_GUI:
                logger.info(
                    '[front] Reselecting renamed selection: %r' % (reselect_new_name,)
                )
        if VERBOSE_GUI:
            logger.info('[front] Reselecting index: %r' % (reselect_index,))
        # ibswgt.species_combo.setOptions(detection_combo_box_options)
        # ibswgt.species_combo.updateOptions(reselect=reselect, reselect_index=reselect_index)

    def setWindowTitle(ibswgt, title):
        parent_ = ibswgt.parent()
        if parent_ is not None:
            parent_.setWindowTitle(title)
        else:
            IBEIS_WIDGET_BASE.setWindowTitle(ibswgt, title)

    def _change_imageset(ibswgt, imgsetid):
        if VERBOSE_GUI:
            logger.info(
                '[newgui] _change_imageset(imgsetid=%r, uuid=%r)'
                % (imgsetid, ibswgt.back.ibs.get_imageset_uuid(imgsetid))
            )
        for tblname in ibswgt.tblname_list:
            view = ibswgt.views[tblname]
            view.clearSelection()
        for tblname in ibswgt.changing_models_gen(tblnames=ibswgt.tblname_list):
            view = ibswgt.views[tblname]
            view._change_imageset(imgsetid)
        try:
            ibswgt.back.select_imgsetid(imgsetid)
        except Exception as ex:
            ut.printex(ex, iswarning=True)
        ibswgt.set_table_tab(IMAGE_TABLE)

    def _update_imageset_tab_name(ibswgt, imgsetid, imagesettext):
        ibswgt.imageset_tabwgt._update_imageset_tab_name(imgsetid, imagesettext)

    # ------------
    # SLOT HELPERS
    # ------------

    def get_table_tab_index(ibswgt, tblname):
        view = ibswgt.views[tblname]
        index = ibswgt._tables_tab_widget.indexOf(view)
        return index

    def set_selection_status(ibswgt, key, ids):
        ids = list(ids)
        text = repr(ids)
        index = ibswgt.tablename_to_status_widget_index[key]

        if len(ids) <= 1:
            text2 = 'Selected %s:' % (ibswgt.key_to_objnice[key],)
        else:
            text2 = 'Selected %s %s:' % (
                len(ids),
                ut.pluralize(ibswgt.key_to_objnice[key], len(ids)),
            )
        ibswgt.status_widget_list[index - 1].setText(text2)
        ibswgt.status_widget_list[index].setText(text)

    def set_table_tab(ibswgt, tblname):
        """
        Programmatically change to the table-tab to either:
        Image, ImageGrid, Annotation, or Names table/tree

        Example:
            >>> # xdoctest: +REQUIRES(--gui)
            >>> from wbia.gui.newgui import *  # NOQA
            >>> ibs, back, ibswgt, testdata_main_loop = testdata_guifront()
            >>> ibswgt.set_table_tab(gh.ANNOTATION_TABLE)
        """
        if VERBOSE_GUI:
            logger.info('[newgui] set_table_tab: %r ' % (tblname,))
        with ut.Timer('set table tab', verbose=VERBOSE_GUI):
            index = ibswgt.get_table_tab_index(tblname)
            ibswgt._tables_tab_widget.setCurrentIndex(index)

    def select_imageset_tab(ibswgt, imgsetid):
        if VERBOSE_GUI:
            logger.info('[newgui] select_imageset_tab imgsetid=%r' % (imgsetid,))
        if isinstance(imgsetid, six.string_types):
            # Hack
            imagesettext = imgsetid
            imgsetid = ibswgt.ibs.get_imageset_imgsetids_from_text(imagesettext)
        else:
            imagesettext = ibswgt.ibs.get_imageset_text(imgsetid)
        # ibswgt.back.select_imgsetid(imgsetid)
        ibswgt.imageset_tabwgt._add_imageset_tab(imgsetid, imagesettext)

    def spawn_edit_image_annotation_interaction_from_aid(
        ibswgt, aid, imgsetid, model=None, qtindex=None
    ):
        """
        hack for letting annots spawn image editing

        CommandLine:
            python -m wbia.gui.newgui spawn_edit_image_annotation_interaction_from_aid --show

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.gui.newgui import *  # NOQA
            >>> import wbia
            >>> main_locals = wbia.main(defaultdb='testdb1')
            >>> ibs, back = ut.dict_take(main_locals, ['ibs', 'back'])
            >>> ibswgt = back.ibswgt  # NOQA
            >>> aid = 4
            >>> imgsetid = 1
            >>> ibswgt.spawn_edit_image_annotation_interaction_from_aid(aid, imgsetid)
            >>> if ut.show_was_requested():
            >>>    gt.qtapp_loop(qwin=ibswgt)
        """
        gid = ibswgt.back.ibs.get_annot_gids(aid)
        if model is None:
            view = ibswgt.views[IMAGE_TABLE]
            model = view.model()
            qtindex, row = view.get_row_and_qtindex_from_id(gid)
        ibswgt.spawn_edit_image_annotation_interaction(model, qtindex, gid, imgsetid)

    def spawn_edit_image_annotation_interaction(ibswgt, model, qtindex, gid, imgsetid):
        """
        TODO: needs reimplement using more standard interaction methods
        """
        logger.info('[newgui] Creating new annotation interaction: gid=%r' % (gid,))
        ibs = ibswgt.ibs
        # Select gid
        ibswgt.back.select_gid(gid, imgsetid, show=False)
        # Interact with gid
        nextcb, prevcb, current_gid = ibswgt._interactannot2_callbacks(model, qtindex)
        iannot2_kw = {
            'rows_updated_callback': ibswgt.update_tables,
            'next_callback': nextcb,
            'prev_callback': prevcb,
        }
        assert current_gid == gid, 'problem in next/prev updater'
        ibswgt.annot_interact = interact_annotations2.ANNOTATION_Interaction2(
            ibs, gid, **iannot2_kw
        )
        # hacky GID_PROG: TODO: FIX WITH OTHER HACKS OF THIS TYPE
        # FIXME; this should depend on the model.
        # _, row = model.view.get_row_and_qtindex_from_id(gid)
        # pt.set_figtitle('%d/%d' % (row + 1, model.rowCount()))
        level_num_rows = model._get_level_row_count(qtindex)
        level_row = model._get_level_row_index(qtindex)
        pt.set_figtitle('%d/%d' % (level_row + 1, level_num_rows))

    def make_adjacent_qtindex_callbacks(ibswgt, model, qtindex):
        r"""
        Returns:
            tuple: (current_rowid, next_callback, prev_callback)

        CommandLine:
            python -m wbia.gui.newgui --test-make_adjacent_qtindex_callbacks

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.gui.newgui import *  # NOQA
            >>> ibs, back, ibswgt, testdata_main_loop = testdata_guifront()
            >>> gid = ibs.get_valid_gids()[0]
            >>> model = ibswgt.models[gh.IMAGE_TABLE]
            >>> qtindex, row = model.get_row_and_qtindex_from_id(gid)
            >>> tup = ibswgt.make_adjacent_qtindex_callbacks(model, qtindex)
            >>> (current_rowid, next_callback, prev_callback) = tup
            >>> assert prev_callback is None, 'should not be a previous image id'
            >>> current_rowid1, next_callback1, prev_callback1 = next_callback()
            >>> assert next_callback() is None, 'race condition not prevented'
            >>> current_rowid2, next_callback2, prev_callback2 = next_callback1()
            >>> testdata_main_loop(globals(), locals())
        """
        current_rowid = model._get_row_id(qtindex)
        next_qtindex = model._get_adjacent_qtindex(qtindex, 1)
        prev_qtindex = model._get_adjacent_qtindex(qtindex, -1)
        next_callback = None
        prev_callback = None
        numclicks = [0]  # semephore, invalidates both functions after one call
        if next_qtindex is not None and next_qtindex.isValid():

            def next_callback():
                if numclicks[0] != 0:
                    logger.info('race condition in next_callback %d ' % numclicks[0])
                    return
                numclicks[0] += 1
                return ibswgt.make_adjacent_qtindex_callbacks(model, next_qtindex)

        if prev_qtindex is not None and prev_qtindex.isValid():

            def prev_callback():
                if numclicks[0] != 0:
                    logger.info('race condition in next_callback %d ' % numclicks[0])
                    return
                numclicks[0] += 1
                return ibswgt.make_adjacent_qtindex_callbacks(model, next_qtindex)

        return current_rowid, next_callback, prev_callback

    def _interactannot2_callbacks(ibswgt, model, qtindex):
        """
        callbacks for the edit image annotation (from image table) interaction

        python -m wbia --db lynx --imgsetid 2

        TODO: needs reimplement
        """
        # if not qtindex.isValid():
        #    raise AssertionError('Bug: qtindex got invalidated')
        #    # BUG: somewhere qtindex gets invalidated
        #    #return None, None, -1
        # HACK FOR NEXT AND PREVIOUS CLICK CALLBACKS
        # logger.info('model.name = %r' % (model.name,))
        if model.name == gh.IMAGE_TABLE:
            cur_gid = model._get_row_id(qtindex)
        # elif model.name == gh.IMAGE_GRID:
        #     cur_gid = model._get_row_id(qtindex)
        elif model.name == gh.NAMES_TREE:
            cur_level = model._get_level(qtindex)
            if cur_level == 1:
                cur_aid = model._get_row_id(qtindex)
                cur_gid = ibswgt.ibs.get_annot_gids(cur_aid)
            else:
                raise NotImplementedError(
                    'Unknown model.name=%r, cur_level=%r' % (model.name, cur_level)
                )
        else:
            logger.info('gh.IMAGE_TABLE = %r' % (gh.IMAGE_TABLE,))
            raise NotImplementedError('Unknown model.name =%r' % (model.name,))
        next_qtindex = model._get_adjacent_qtindex(qtindex, 1)
        prev_qtindex = model._get_adjacent_qtindex(qtindex, -1)
        numclicks = [0]  # semephore

        def make_qtindex_callback(qtindex_, type_='nextprev'):
            def _qtindex_callback():
                if numclicks[0] != 0:
                    logger.info(
                        'race condition in %s_callback %d ' % (type_, numclicks[0])
                    )
                    return
                numclicks[0] += 1
                # call this function again with next index
                nextcb, prevcb, new_gid1 = ibswgt._interactannot2_callbacks(
                    model, qtindex_
                )
                logger.info('[newgui] %s_callback: new_gid1=%r' % (type_, new_gid1))
                ibswgt.annot_interact.update_image_and_callbacks(
                    new_gid1, nextcb, prevcb, do_save=True
                )
                # hacky GID_PROG: TODO: FIX WITH OTHER HACKS OF THIS TYPE
                # _, row = model.view.get_row_and_qtindex_from_id(new_gid1)
                # pt.set_figtitle('%d/%d' % (row + 1, model.rowCount()))
                level_num_rows = model._get_level_row_count(qtindex_)
                level_row = model._get_level_row_index(qtindex_)
                pt.set_figtitle('%d/%d' % (level_row + 1, level_num_rows))

            return _qtindex_callback

        if next_qtindex is not None and next_qtindex.isValid():
            next_callback = make_qtindex_callback(next_qtindex, 'next')
        else:
            next_callback = None

        if prev_qtindex is not None and prev_qtindex.isValid():
            prev_callback = make_qtindex_callback(prev_qtindex, 'prev')
        else:
            prev_callback = None

        return next_callback, prev_callback, cur_gid

    # ------------
    # SLOTS
    # ------------

    @slot_()
    def select_annot_text_editing_finished(ibswgt):
        tablename = gh.ANNOTATION_TABLE
        index = ibswgt.tablename_to_status_widget_index[tablename]
        text = ibswgt.status_widget_list[index].text()
        ibswgt.select_table_indicies_from_text(tablename, text)

    @slot_()
    def select_name_text_editing_finished(ibswgt):
        tablename = gh.NAMES_TREE
        index = ibswgt.tablename_to_status_widget_index[tablename]
        text = ibswgt.status_widget_list[index].text()
        ibswgt.select_table_indicies_from_text(tablename, text)

    @slot_()
    def select_image_text_editing_finished(ibswgt):
        tablename = gh.IMAGE_TABLE
        index = ibswgt.tablename_to_status_widget_index[tablename]
        text = ibswgt.status_widget_list[index].text()
        ibswgt.select_table_indicies_from_text(tablename, text)

    def select_table_indicies_from_text(ibswgt, tblname, text, allow_table_change=False):
        """
        Args:
            tblname - tablename of the id to parse from text

        Ignore:
            text = '[1, 2,  3,]'
               text = '51e10019-968b-5f2e-2287-8432464d7547 '

        Example:
            >>> # xdoctest: +REQUIRES(--gui)
            >>> from wbia.gui.newgui import *  # NOQA
            >>> ibs, back, ibswgt, testdata_main_loop = testdata_guifront()
            >>> ibswgt.set_table_tab(gh.ANNOTATION_TABLE)
            >>> tblname = gh.NAMES_TREE
            >>> text = 'lena'
            >>> ibswgt.select_table_indicies_from_text(tblname, text)
        """
        if not ut.QUIET:
            logger.info('[newgui] select_table_indicies_from_text')
            logger.info('[newgui]  * gh.tblname = %r' % (tblname,))
            logger.info('[newgui]  * text = %r' % (text,))
        to_backend_tablename = {
            gh.ANNOTATION_TABLE: const.ANNOTATION_TABLE,
            gh.NAMES_TREE: const.NAME_TABLE,
            gh.IMAGE_TABLE: const.IMAGE_TABLE,
        }
        backend_tablename = to_backend_tablename[tblname]
        if not ut.QUIET:
            logger.info('[newgui]  * backend_tablename = %r' % (backend_tablename,))
        if text == '':
            text = '[]'
        try:
            # MODE1 = True
            # if MODE1:
            id_list_ = text.lstrip('[').rstrip(']').split(',')
            id_list = [id_.strip() for id_ in id_list_]
            id_list = [id_ for id_ in id_list if len(id_) > 0]
            try:
                id_list = list(map(int, id_list))
            except ValueError:
                import uuid

                try:
                    # First check to see if the text is a UUID
                    id_list = list(map(uuid.UUID, id_list))
                except ValueError:
                    if tblname != gh.NAMES_TREE:
                        raise
                    else:
                        # then maybe it was a name that was selected
                        id_list = list(map(str, id_list))
            # else:
            #    id_list_ = eval(text, globals(), locals())
            #    id_list = ut.ensure_iterable(id_list_)  # NOQA
        except Exception as ex:
            ut.printex(ex, iswarning=True, keys=['text'])
        else:
            if not ut.QUIET:
                logger.info('[newgui]  * id_list = %r' % (id_list,))
            # logger.info(id_list)
            id_list = ibswgt.back._set_selection3(backend_tablename, id_list, mode='set')

        # Select the index if we are in the right table tab
        if len(id_list) == 1 and (
            allow_table_change or ibswgt._tables_tab_widget.current_tblname == tblname
        ):
            if not ut.QUIET:
                logger.info('[newgui]  * attempting to select from rowid')
            # view = ibswgt.views[tblname]
            # view.select_row_from_id(id_list[0])
            ibswgt.goto_table_id(tblname, id_list[0])
        else:
            # TODO: convert the id into the ids corresponding with this tablename and move
            # to the first one
            pass

        ibswgt.back.update_selection_texts()
        # pass

    @slot_(str, int)
    def on_rows_updated(ibswgt, tblname, nRows):
        """
        When the rows are updated change the tab names
        """
        if VERBOSE_GUI:
            logger.info(
                '[newgui] on_rows_updated: tblname=%12r nRows=%r ' % (tblname, nRows)
            )
        if tblname == IMAGESET_TABLE:  # Hack
            # logger.info('... tblname == IMAGESET_TABLE, ...hack return')
            return
        tblname = str(tblname)
        TABLE_NICE = ibswgt.declare_tup[1]  # hack
        tblnice = TABLE_NICE[tblname]
        index = ibswgt.get_table_tab_index(tblname)
        text = tblnice + ' ' + str(nRows)
        # CHANGE TAB NAME TO SHOW NUMBER OF ROWS
        ibswgt._tables_tab_widget.setTabText(index, text)

    def goto_table_id(ibswgt, tablename, _id):
        logger.info('[newgui] goto_table_id(tablenamd=%r, _id=%r)' % (tablename, _id))
        ibswgt.set_table_tab(tablename)
        view = ibswgt.views[tablename]
        view.select_row_from_id(_id, scroll=True)

    @slot_(QtCore.QModelIndex, QtCore.QPoint)
    def on_contextMenuClicked(ibswgt, qtindex, pos):
        """
        Right click anywhere in the GUI
        Context menus on right click of a table

        CommandLine:
            python -m wbia --db WS_ALL --imgsetid 2 --select-name=A-003
        """
        if not qtindex.isValid():
            return

        model = qtindex.model()
        tblview = ibswgt.views[model.name]
        context_options = []
        qtindex_list = tblview.selectedIndexes()
        id_list = [model._get_row_id(_qtindex) for _qtindex in qtindex_list]
        level_list = [model._get_level(_qtindex) for _qtindex in qtindex_list]
        level2_ids_ = ut.group_items(id_list, level_list)
        level2_ids = {
            level: ut.unique_ordered(ids) for level, ids in six.iteritems(level2_ids_)
        }

        ibs = ibswgt.back.ibs
        back = ibswgt.back

        col = qtindex.column()
        setter_func = model.col_setter_list[col]
        if setter_func is not None:
            # Make a copy of the list
            id_list_ = id_list[:]
            clone_highlight = len(id_list_) > 1
            if not clone_highlight:
                id_list_ = [_.get_id() for _ in model.level_index_list]
            # Getter and current value
            getter_func = model.col_getter_list[col]
            current_id = model._get_row_id(qtindex)
            current_value = getter_func(current_id)
            current_list = [current_value] * len(id_list_)
            # Add to context menus

            def _setter_refresh(id_list_, current_list, tablename):
                args = (
                    current_list[0],
                    len(id_list_),
                )
                if not back.are_you_sure('Confirm cloning value %s to %d rows?' % args):
                    return
                setter_func(id_list_, current_list)
                ibswgt.update_tables(tblnames=[tablename], clear_view_selection=False)

            if clone_highlight:
                context_options += [
                    (
                        'Clone selected value to highlighted rows',
                        lambda: _setter_refresh(id_list_, current_list, model.name),
                    ),
                    ('----', lambda: None),
                ]
            else:
                context_options += [
                    (
                        'Clone selected value to all rows',
                        lambda: _setter_refresh(id_list_, current_list, model.name),
                    ),
                    ('----', lambda: None),
                ]

        def build_annot_context_options(ibswgt, ibs, aid_list, imgsetid, **kwargs):
            context_options = []
            if len(aid_list) == 1:
                aid = aid_list[0]
                if kwargs.get('goto_image', True):
                    context_options += [
                        (
                            'Go to image',
                            lambda: ibswgt.goto_table_id(
                                IMAGE_TABLE, ibswgt.back.ibs.get_annot_gids(aid)
                            ),
                        )
                    ]
                if kwargs.get('goto_annot', True):
                    context_options += [
                        (
                            'Go to annot',
                            lambda: ibswgt.goto_table_id(gh.ANNOTATION_TABLE, aid),
                        )
                    ]
                if kwargs.get('goto_name', True):
                    context_options += [
                        (
                            'Go to name',
                            lambda: ibswgt.goto_table_id(
                                NAMES_TREE, ibswgt.back.ibs.get_annot_nids(aid)
                            ),
                        ),
                    ]
                if kwargs.get('canedit', True):
                    context_options += [
                        (
                            'Edit Annotation in Image',
                            lambda: ibswgt.spawn_edit_image_annotation_interaction_from_aid(
                                aid, imgsetid
                            ),
                        ),
                    ]
                    context_options += [
                        (
                            "Clone annotation's species to ImageSet",
                            lambda: ibswgt.back.copy_species_to_imageset(aid, imgsetid),
                        ),
                    ]

                context_options += [
                    ('----', lambda: None),
                    (
                        'View annotation in Web',
                        # lambda: ibswgt.back.select_aid(aid, imgsetid, show=True)),
                        lambda: ibswgt.back.show_annotation(aid, web=True),
                    ),
                    (
                        'View image in Web',
                        lambda: ibswgt.back.select_gid_from_aid(
                            aid, imgsetid, show=True, web=True
                        ),
                    ),
                    (
                        'Set annotation species',
                        lambda: ibswgt.back.override_all_annotation_species(aid_list),
                    ),
                    ('----', lambda: None),
                    ("Remove annotation's name", lambda: ibswgt.back.unset_names([aid])),
                    ('Delete annotation', lambda: ibswgt.back.delete_annot(aid_list)),
                    ('----', lambda: None),
                ]
                from wbia.viz.interact import interact_chip
                from wbia import viz

                context_options += interact_chip.build_annot_context_options(
                    ibswgt.back.ibs,
                    aid,
                    refresh_func=viz.draw,
                    with_interact_image=False,
                )
            else:
                context_options += [
                    (
                        'View annotations in Web',
                        lambda: ibswgt.back.show_aid_list_in_web(aid_list),
                    ),
                    (
                        'Set annotation species',
                        lambda: ibswgt.back.override_all_annotation_species(aid_list),
                    ),
                    (
                        "Unset annotations' names",
                        lambda: ibswgt.back.unset_names(aid_list),
                    ),
                    ('Delete annotations', lambda: ibswgt.back.delete_annot(aid_list)),
                ]
            return context_options

        def name_context_options(ibswgt, ibs, nid_list, aid_list, imgsetid, **kwargs):
            context_options = []
            if len(aid_list) == 1:
                aid = aid_list[0]
                context_options += build_annot_context_options(
                    ibswgt, ibs, [aid], imgsetid
                )
            if len(aid_list) > 0:

                def set_annot_names_to_same_new_name(ibswgt, aid_list):
                    ibswgt.back.ibs.set_annot_names_to_same_new_name(aid_list)
                    ibswgt.update_tables(tblnames=[gh.NAMES_TREE])

                context_options += [
                    (
                        'Rename annots (%s) to new name'
                        % ut.list_str_summarized(aid_list, 'aid_list'),
                        lambda: set_annot_names_to_same_new_name(ibswgt, aid_list),
                    ),
                ]
            if len(nid_list) > 0:

                def run_splits(ibs, nid_list):
                    logger.info('Checking for splits')
                    aids_list = ibs.get_name_aids(nid_list)
                    aid_list = sorted(list(set(ut.flatten(aids_list))))
                    back.run_annot_splits(aid_list)

                def export_nids(ibs, nid_list):
                    from wbia.dbio import export_subset

                    if not back.are_you_sure(
                        'Confirm export of nid_list=%r' % (nid_list,)
                    ):
                        return
                    export_subset.export_names(ibs, nid_list)

                def create_new_imageset_from_names_(ibs, nid_list):
                    ibs.create_new_imageset_from_names(nid_list)
                    ibswgt.update_tables([gh.IMAGESET_TABLE], clear_view_selection=False)

                context_options += [
                    (
                        'View name(s) in Web',
                        lambda: ibswgt.back.show_nid_list_in_web(nid_list),
                    ),
                    ('----', lambda: None),
                    ('Check for splits', lambda: run_splits(ibs, nid_list)),
                    ('Export names', lambda: export_nids(ibs, nid_list)),
                    (
                        'Create ImageSet From Name(s)',
                        lambda: create_new_imageset_from_names_(ibs, nid_list),
                    ),
                ]

                from wbia.viz.interact import interact_name

                context_options += interact_name.build_name_context_options(
                    ibswgt.back.ibs, nid_list
                )
            elif len(nid_list) == 0:
                # from wbia.viz.interact import interact_name
                # context_options += interact_name.build_name_context_options(
                #    ibswgt.back.ibs, nid_list)
                # logger.info('nutin')
                pass
            return context_options

        def build_image_context_options(ibswgt, ibs, gid_list, imgsetid, **kwargs):
            current_imagesettext = ibswgt.back.ibs.get_imageset_text(imgsetid)
            context_options = []
            # Conditional context menu

            def view_images_in_directory(gid_list):
                images = ibs.images(gid_list)
                fpaths = images.paths
                ut.view_file_in_directory(fpaths)

            def view_images_in_web(gid_list):
                ibswgt.back.show_images_in_web(gid_list)

            fmt = {
                'images': ut.pluralize('image', len(gid_list)),
                "images's": ut.pluralize("image's", len(gid_list)),
                'annotations': ut.pluralize('annotation', len(gid_list)),
                'times': ut.pluralize('time', len(gid_list)),
            }

            context_options = [
                (
                    "Edit {images's} {times}".format(**fmt),
                    lambda: ibswgt.edit_image_time(gid_list),
                ),
                (
                    'View {images} in Web'.format(**fmt),
                    lambda: view_images_in_web(gid_list),
                ),
                (
                    'View {images} in Directory'.format(**fmt),
                    lambda: view_images_in_directory(gid_list),
                ),
                ('---', lambda: None),
                (
                    'Add annotation from entire {images}'.format(**fmt),
                    lambda: ibswgt.back.add_annotation_from_image(gid_list),
                ),
                (
                    'Run detection on {images} (can cause duplicates)'.format(**fmt),
                    lambda: ibswgt.back.run_detection_on_images(gid_list),
                ),
            ]

            if len(gid_list) == 1:
                gid = gid_list[0]
                imgsetid = model.imgsetid
                aid_list = ibs.get_image_aids(gid)

                annot_options = [
                    (
                        'Options aid=%r' % (aid,),
                        build_annot_context_options(
                            ibswgt, ibs, [aid], imgsetid, goto_image=False
                        ),
                    )
                    for aid in aid_list
                ]
                if len(aid_list) == 1:
                    annot_option_item = (
                        'Annot Options (aid=%r)' % (aid_list[0],),
                        annot_options[0][1],
                    )
                else:
                    annot_option_item = ('Annot Options', annot_options)
                if kwargs.get('goto_image_in_imgtbl', False):
                    context_options += [
                        (
                            'Go to image in Images Table',
                            lambda: ibswgt.goto_table_id(IMAGE_TABLE, gid),
                        ),
                    ]

                context_options += [
                    (
                        'View image in Matplotlib',
                        lambda: ibswgt.back.select_gid(
                            gid, imgsetid, show=True, web=False
                        ),
                    ),
                    (
                        'View detection image (Hough) [dev]',
                        lambda: ibswgt.back.show_hough_image_(gid),
                    ),
                    annot_option_item,
                ]
            # Special condition for imagesets
            if current_imagesettext != const.NEW_IMAGESET_IMAGESETTEXT:
                context_options += [
                    ('----', lambda: None),
                    (
                        'Move to new imageset',
                        lambda: ibswgt.back.send_to_new_imageset(gid_list, mode='move'),
                    ),
                    (
                        'Copy to new imageset',
                        lambda: ibswgt.back.send_to_new_imageset(gid_list, mode='copy'),
                    ),
                ]
            if current_imagesettext != const.UNGROUPED_IMAGES_IMAGESETTEXT:
                context_options += [
                    ('----', lambda: None),
                    (
                        'Remove from imageset',
                        lambda: ibswgt.back.remove_from_imageset(gid_list),
                    ),
                ]

            from wbia.viz import viz_graph2

            context_options += [
                ('----', lambda: None),
                (
                    'Graph interaction (Annots)',
                    lambda: viz_graph2.make_qt_graph_interface(
                        # ibs, nids=ut.unique(ut.flatten(ibs.get_image_nids(gid_list))),
                        ibs,
                        gids=gid_list,
                        graph_tab=True,
                    ),
                ),
            ]

            # Continue the conditional context menu
            if len(gid_list) == 1:
                # We get gid from above
                context_options += [
                    ('----', lambda: None),
                    (
                        "Delete image's annotations",
                        lambda: ibswgt.back.delete_image_annotations([gid]),
                    ),
                    (
                        "Set image's annotations species",
                        lambda: ibswgt.back.override_all_annotation_species(gids=[gid]),
                    ),
                    ('Delete image', lambda: ibswgt.back.delete_image(gid)),
                ]
            else:
                context_options += [
                    ('----', lambda: None),
                    (
                        "Delete images' annotations",
                        lambda: ibswgt.back.delete_image_annotations(gid_list),
                    ),
                    (
                        "Set images' annotations species",
                        lambda: ibswgt.back.override_all_annotation_species(
                            gids=gid_list
                        ),
                    ),
                    ('Delete images', lambda: ibswgt.back.delete_image(gid_list)),
                ]
            return context_options

        # ---- IMAGESET CONTEXT ----
        if model.name == IMAGESET_TABLE:
            # This is for the benefit of merge imagesets
            merge_destination_id = model._get_row_id(qtindex)
            imagesettext = ibswgt.back.ibs.get_imageset_text(merge_destination_id)
            imgsetid_list = level2_ids[0]
            # Conditional context menu
            # TODO: remove duplicate code
            if len(imgsetid_list) == 1:
                context_options += [
                    (
                        'View imageset in Web',
                        lambda: ibswgt.back.show_imgsetid_list_in_web(imgsetid_list),
                    ),
                    (
                        "Turk imageset's detections in Web",
                        lambda: ibswgt.back.show_imgsetid_detection_turk_in_web(
                            imgsetid_list
                        ),
                    ),
                    (
                        "Turk imageset's annotations in Web",
                        lambda: ibswgt.back.show_imgsetid_annotation_turk_in_web(
                            imgsetid_list
                        ),
                    ),
                    ('----', lambda: None),
                    (
                        'Run detection on imageset (can cause duplicates)',
                        lambda: ibswgt.back.run_detection_on_imageset(imgsetid_list),
                    ),
                    (
                        'Merge %d imageset into %s'
                        % (len(imgsetid_list), (imagesettext)),
                        lambda: ibswgt.back.merge_imagesets(
                            imgsetid_list, merge_destination_id
                        ),
                    ),
                    ('Copy imageset', lambda: ibswgt.back.copy_imageset(imgsetid_list)),
                    (
                        'Export imageset',
                        lambda: ibswgt.back.export_imagesets(imgsetid_list),
                    ),
                    ('----', lambda: None),
                    (
                        'Delete imageset',
                        lambda: ibswgt.back.delete_imageset(imgsetid_list),
                    ),
                    ('----', lambda: None),
                    (
                        'Delete imageset AND images',
                        lambda: ibswgt.back.delete_imageset_and_images(imgsetid_list),
                    ),
                    ('----', lambda: None),
                    (
                        'Mark imageset as Shipped to WildBook',
                        lambda: ibswgt.back.mark_imageset_as_shipped(imgsetid_list),
                    ),
                ]
            else:
                context_options += [
                    (
                        'Run detection on imagesets (can cause duplicates)',
                        lambda: ibswgt.back.run_detection_on_imageset(imgsetid_list),
                    ),
                    ('Copy imageset', lambda: ibswgt.back.copy_imageset(imgsetid_list)),
                    (
                        'Merge %d imagesets into %s'
                        % (len(imgsetid_list), (imagesettext)),
                        lambda: ibswgt.back.merge_imagesets(
                            imgsetid_list, merge_destination_id
                        ),
                    ),
                    ('----', lambda: None),
                    (
                        'Delete imagesets',
                        lambda: ibswgt.back.delete_imageset(imgsetid_list),
                    ),
                    ('----', lambda: None),
                    (
                        'Delete imagesets AND images',
                        lambda: ibswgt.back.delete_imageset_and_images(imgsetid_list),
                    ),
                    # ('export imagesets', lambda: ibswgt.back.export_imagesets(imgsetid_list)),
                ]
        # ---- IMAGE CONTEXT ----
        elif model.name == IMAGE_TABLE:
            gid_list = level2_ids[0]
            imgsetid = ibswgt.back.get_selected_imgsetid()
            context_options += build_image_context_options(
                ibswgt, ibs, gid_list, imgsetid
            )
        # ---- IMAGE GRID CONTEXT ----
        elif model.name == IMAGE_GRID:
            gid_list = level2_ids[0]
            imgsetid = ibswgt.back.get_selected_imgsetid()
            context_options += build_image_context_options(
                ibswgt, ibs, gid_list, imgsetid, goto_image_in_imgtbl=True
            )
        # ---- ANNOTATION CONTEXT ----
        elif model.name == gh.ANNOTATION_TABLE:
            aid_list = level2_ids[0]
            # Conditional context menu
            # TODO: UNIFY COMMMON CONTEXT MENUS
            context_options += build_annot_context_options(
                ibswgt, ibs, aid_list, model.imgsetid, goto_annot=False
            )
        # ---- NAMES TREE CONTEXT ----
        elif model.name == NAMES_TREE:
            # TODO: map level list to tablename more reliably
            ut.print_dict(level2_ids)
            nid_list = level2_ids.get(0, [])
            aid_list = level2_ids.get(1, [])
            if len(aid_list) > 0 and len(nid_list) > 0:
                # two types of indices are selected, just return
                # fixme to do something useful
                logger.info('multiple types of indicies selected')
                return
            else:
                imgsetid = model.imgsetid
                context_options += name_context_options(
                    ibswgt, ibs, nid_list, aid_list, imgsetid
                )
        # Show the context menu
        # ut.print_list(context_options, nl=2)
        if len(context_options) > 0:
            gt.popup_menu(tblview, pos, context_options)

    @slot_(QtCore.QModelIndex)
    def on_doubleclick(ibswgt, qtindex):
        """
        Double clicking anywhere in the GUI

        CommandLine:
            python -m wbia --db lynx --imgsetid 2 --select-name=goku
        """
        logger.info('\n+--- DOUBLE CLICK ---')
        if not qtindex.isValid():
            logger.info('[doubleclick] invalid qtindex')
            return
        model = qtindex.model()
        id_ = model._get_row_id(qtindex)
        if model.name == IMAGESET_TABLE:
            imgsetid = id_
            ibswgt.select_imageset_tab(imgsetid)
        else:
            imgsetid = model.imgsetid
            if (model.name == IMAGE_TABLE) or (model.name == IMAGE_GRID):
                gid = id_
                ibswgt.spawn_edit_image_annotation_interaction(
                    model, qtindex, gid, imgsetid
                )
            elif model.name == gh.ANNOTATION_TABLE:
                aid = id_
                ibswgt.back.select_aid(aid, imgsetid)
            elif model.name == NAME_TABLE:
                nid = id_
                ibswgt.back.select_nid(nid, imgsetid)
            elif model.name == NAMES_TREE:
                level = model._get_level(qtindex)
                if level == 0:
                    nid = id_
                    ibswgt.back.select_nid(nid, imgsetid, show=True)
                elif level == 1:
                    aid = id_
                    ibswgt.spawn_edit_image_annotation_interaction_from_aid(
                        aid, imgsetid, model, qtindex
                    )
                    # ibswgt.back.select_aid(aid, imgsetid, show=True)

    # @slot_(list)
    def imagesDropped(ibswgt, url_list):
        r"""
        image drag and drop event

        CommandLine:
            python -m wbia.gui.newgui imagesDropped --show

        Example:
            >>> # xdoctest: +REQUIRES(--gui)
            >>> from wbia.gui.newgui import *  # NOQA
            >>> ibs, back, ibswgt, testdata_main_loop = testdata_guifront('hstest')
            >>> url_list = [ut.grab_test_imgpath('carl.jpg'), ut.grab_test_imgpath('lena.png')]
            >>> #url_list += [ut.truepath('~/Downloads/Clutter/wd_peter2.zip')]
            >>> url = url_list[0]
            >>> ut.quit_if_noshow()
            >>> ibswgt.imagesDropped(url_list)
            >>> testdata_main_loop(globals(), locals())
        """
        logger.info('[drop_event] url_list=%r' % (url_list,))
        has_zipext = ut.partial(ut.fpath_has_ext, exts=['.zip'])
        gpath_list = list(filter(ut.fpath_has_imgext, url_list))
        dir_list = list(filter(isdir, url_list))
        zipfile_list = list(filter(has_zipext, url_list))
        old = False
        if old:
            if len(dir_list) > 0:
                options = ['No', 'Yes']
                title = 'Non-Images dropped'
                msg = 'Recursively import from directories?'
                ans = gt.user_option(ibswgt, msg=msg, title=title, options=options)
                if ans == 'Yes':
                    unflat_gpaths = [
                        ut.list_images(dir_, fullpath=True, recursive=True)
                        for dir_ in dir_list
                    ]
                    flat_gpaths = ut.flatten(unflat_gpaths)
                    flat_unix_gpaths = list(map(ut.unixpath, flat_gpaths))
                    gpath_list.extend(flat_unix_gpaths)
                else:
                    return
            logger.info('[drop_event] gpath_list=%r' % (gpath_list,))
            if len(gpath_list) > 0:
                ibswgt.back.import_images_from_file(gpath_list=gpath_list)
        else:
            from wbia.dbio import ingest_database

            ibs = ibswgt.back.ibs
            ingestable = ingest_database.Ingestable2(
                ibs.get_dbdir(), gpath_list, dir_list, zipfile_list
            )
            num_gpaths = len(ingestable.imgpath_list)
            num_dpaths = len(ingestable.imgdir_list)
            num_zips = len(ingestable.zipfile_list)
            confirm_list = []
            if num_gpaths > 0:
                confirm_list += [ut.quantstr('image file', num_gpaths)]
            if num_dpaths > 0:
                confirm_list += [
                    'recursively from ' + ut.quantstr('directory', num_dpaths, 's')
                ]
            if num_zips > 0:
                confirm_list += [ut.quantstr('zip file', num_zips, 's')]
            confirm_msg = 'Import from: ' + ut.conj_phrase(confirm_list, 'and') + '.'
            # gt.rrrr()
            config = ingestable.ingest_config
            # cfg = config
            dlg = gt.ConfigConfirmWidget.as_dialog(
                ibswgt, title='Confirm Import Images', msg=confirm_msg, config=config
            )
            dlg.resize(700, 500)
            self = dlg.widget
            dlg.exec_()
            logger.info('config = %r' % (config,))
            updated_config = self.config  # NOQA
            logger.info('updated_config = %r' % (updated_config,))
            gid_list = ingestable.execute(ibs=ibs)
            ibswgt.back._process_new_images(
                refresh=True, gid_list=gid_list, clock_offset=False
            )

    def select_table_id(ibswgt, table_key, level, id_, imgsetid):
        select_func_dict = {
            (IMAGE_TABLE, 0): ibswgt.back.select_gid,
            (IMAGE_GRID, 0): ibswgt.back.select_gid,
            (gh.ANNOTATION_TABLE, 0): ibswgt.back.select_aid,
            (NAME_TABLE, 0): ibswgt.back.select_nid,
            (NAMES_TREE, 0): ibswgt.back.select_nid,
            (NAMES_TREE, 1): ibswgt.back.select_aid,
        }
        select_func = select_func_dict[(table_key, level)]
        select_func(id_, imgsetid, show=False)

    def edit_image_time(ibswgt, gid_list):
        """

        CommandLine:
            python -m wbia.gui.newgui --exec-edit_image_time --show

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.gui.newgui import *  # NOQA
            >>> ibs, back, ibswgt, testdata_main_loop = testdata_guifront('testdb3')
            >>> #ibs, back, ibswgt, testdata_main_loop = testdata_guifront('lynx')
            >>> ibswgt.edit_image_time([277, 630])
            >>> testdata_main_loop(globals(), locals())
        """
        from wbia.gui import clock_offset_gui

        ibswgt.co_wgt = clock_offset_gui.ClockOffsetWidget(
            ibswgt.ibs, gid_list, hack=True
        )
        ibswgt.co_wgt.show()

    def filter_annotation_table(ibswgt):
        r"""
        TODO:  Finish implementation

        CommandLine:
            python -m wbia.gui.newgui --test-filter_annotation_table --show --db lynx --imgsetid 2

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.gui.newgui import *  # NOQA
            >>> ibs, back, ibswgt, testdata_main_loop = testdata_guifront('testdb3')
            >>> #ibs, back, ibswgt, testdata_main_loop = testdata_guifront('PZ_Master1')
            >>> result = ibswgt.filter_annotation_table()
            >>> print(result)
            >>> testdata_main_loop(globals(), locals())
        """
        from functools import partial

        # ibs.filter_annots_general()

        ibs = ibswgt.back.ibs

        # ibs.filterannots_by_tags(aid_list)
        logger.info('\n------FILTERING ANNOTS\n\n')

        # annotmatch_rowid_list = ibs._get_all_annotmatch_rowids()
        # isscenerymatch_list = ibs.get_annotmatch_is_scenerymatch(annotmatch_rowid_list)
        # ut.take(isscenerymatch_list, ut.list_where(isscenerymatch_list))

        # Applies annotation based filtering to the annotation table
        # filter_kw = dict(any_matches='.*error.*', been_adjusted=True)
        filter_kw = dict(been_adjusted=True)
        # filter_kw = dict(require_timestamp=True)
        filter_fn = partial(ibs.filter_annots_general, filter_kw=filter_kw)

        model = ibswgt.models[gh.ANNOTATION_TABLE]  # NOQA
        model.set_ider_filters([filter_fn])
        with ChangeLayoutContext([model]):
            model._update_rows(rebuild_structure=True)


# #####################
# ##### Testing #######
# #####################


def testdata_guifront(defaultdb='testdb1'):
    import wbia

    main_locals = wbia.main(defaultdb=defaultdb, allow_newdir=True)
    ibs, back = ut.dict_take(main_locals, ['ibs', 'back'])
    ibswgt = back.ibswgt  # NOQA
    globals__ = globals()
    locals__ = locals()

    def testdata_main_loop(globals_=globals__, locals_=locals__):
        locals_ = locals_.copy()
        globals_ = globals_.copy()
        locals_.update(locals__)
        globals_.update(globals__)
        if '--cmd' in sys.argv:
            gt.qtapp_loop(qwin=ibswgt, ipy=True)
            six.exec_(ut.ipython_execstr(), globals_, locals_)
        elif ut.show_was_requested():
            gt.qtapp_loop(qwin=ibswgt)

    return ibs, back, ibswgt, testdata_main_loop


def testfunc():
    r"""
    CommandLine:
        python -m wbia.gui.newgui --test-testfunc --show
        python -m wbia.gui.newgui --test-testfunc --cmd

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.gui.newgui import *  # NOQA
        >>> result = testfunc()
        >>> print(result)
    """
    ibs, back, ibswgt, testdata_main_loop = testdata_guifront()
    view = ibswgt.views[gh.IMAGE_TABLE]
    testdata_main_loop(globals(), locals())



# -*- coding: utf-8 -*-
import utool as ut
from wbia import viz
from wbia.viz import viz_helpers as vh
from wbia.plottool import draw_func2 as df2
from wbia.plottool import interact_helpers as ih

(print, print_, printDBG, rrr, profile) = ut.inject(
    __name__, '[interact_img]', DEBUG=False
)


# @ut.indent_func
def ishow_image(ibs, gid, sel_aids=[], fnum=None, select_callback=None, **kwargs):
    if ut.VERBOSE:
        print(ut.get_caller_name(range(9)))
        print('[interact_image] gid=%r fnum=%r' % (gid, fnum))
    if fnum is None:
        fnum = df2.next_fnum()
    # TODO: change to class based structure
    self = ut.DynStruct()
    self.fnum = fnum

    fig = ih.begin_interaction('image', fnum)
    # printDBG(utool.func_str(interact_image, [], locals()))
    kwargs['draw_lbls'] = kwargs.get('draw_lbls', True)

    def _image_view(sel_aids=sel_aids, **_kwargs):
        try:
            viz.show_image(ibs, gid, sel_aids=sel_aids, fnum=self.fnum, **_kwargs)
            df2.set_figtitle('Image View')
        except TypeError as ex:
            ut.printex(ex, ut.repr2(_kwargs))
            raise

    # Create callback wrapper
    def _on_image_click(event):
        print('[inter] clicked image')
        if ih.clicked_outside_axis(event):
            # Toggle draw lbls
            kwargs['draw_lbls'] = not kwargs.get('draw_lbls', True)
            _image_view(**kwargs)
        else:
            ax = event.inaxes
            viztype = vh.get_ibsdat(ax, 'viztype')
            annotation_centers = vh.get_ibsdat(ax, 'annotation_centers', default=[])
            print(' annotation_centers=%r' % annotation_centers)
            print(' viztype=%r' % viztype)
            if len(annotation_centers) == 0:
                print(' ...no chips exist to click')
                return
            x, y = event.xdata, event.ydata
            # Find ANNOTATION center nearest to the clicked point
            aid_list = vh.get_ibsdat(ax, 'aid_list', default=[])
            import vtool as vt

            centx, _dist = vt.nearest_point(x, y, annotation_centers)
            aid = aid_list[centx]
            print(' ...clicked aid=%r' % aid)
            if select_callback is not None:
                # HACK, should just implement this correctly here
                select_callback(gid, sel_aids=[aid], fnum=self.fnum)
            else:
                _image_view(sel_aids=[aid])

        viz.draw()

    _image_view(**kwargs)
    viz.draw()
    ih.connect_callback(fig, 'button_press_event', _on_image_click)



# -*- coding: utf-8 -*-
import logging
import six
from six.moves import zip, map
import numpy as np
import vtool as vt
import utool as ut
from wbia.control import controller_inject

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


# Create dectorator to inject functions in this module into the IBEISController
CLASS_INJECT_KEY, register_ibs_method = controller_inject.make_ibs_register_decorator(
    __name__
)


# TODO : make a annot_tags file


ANNOTMATCH_PROPS_STANDARD = [
    # 'SceneryMatch',
    # 'Photobomb',
    # 'Hard',
    # 'NonDistinct',
]

ANNOTMATCH_PROPS_OTHER = [
    'SceneryMatch',
    'Photobomb',
    'Hard',
    'NonDistinct',
    'Occlusion',
    'Viewpoint',
    'MildViewpoint',
    'Pose',
    'Lighting',
    'Quality',  # quality causes failure
    'Orientation',  # orientation caused failure
    'EdgeMatch',  # descriptors on the edge of the naimal produce strong matches
    'Interesting',  # flag a case as interesting
    'JoinCase',  # case should actually be marked as correct
    'SplitCase',  # case should actually be marked as correct
    'random',  # gf case has random matches, the gt is to blame
    'BadShoulder',  # gf is a bad shoulder match
    'BadTail',  # gf is a bad tail match
    'TimeDeltaError',
    # These annots have almost the same information
    'NearDuplicate',
    'CorrectPhotobomb',  # FIXME: this is a terrible name
]

OLD_ANNOTMATCH_PROPS = [
    'TooLargeMatches',  # really big nondistinct matches
    'TooSmallMatches',  # really big nondistinct matches
    'ScoringIssue',  # matches should be scored differently
    'BadCoverage',  # matches were not in good places (missing matches)
    'ViewpointOMG',  # gf is a bad tail match
    'ViewpointCanDo',  # gf is a bad tail match
    'shouldhavemore',
    'Shadowing',  # shadow causes failure
    'success',  # A good success case
    'GoodCoverage',  # matches were spread out correctly (scoring may be off though)
]

# Changes to prop names
PROP_MAPPING = {
    'ViewpointCanDo': 'Correctable',
    'ViewpointOMG': 'Uncorrectable',
    'Shadowing': 'Lighting',
    'success': None,
    'GoodCoverage': None,
    # 'Hard'           : 'NeedsWork',
    'shouldhavemore': 'NeedsWork',
    'BadCoverage': 'NeedsWork',
    'ScoringIssue': 'NeedsWork',
    'TooSmallMatches': 'FeatureScale',
    'TooLargeMatches': 'FeatureScale',
    # 'BadShoulder' : 'BadShoulder',
    # 'GoodCoverage': None,
}

for key, val in PROP_MAPPING.items():
    if key in ANNOTMATCH_PROPS_OTHER:
        ANNOTMATCH_PROPS_OTHER.remove(key)
    if val is not None and val not in ANNOTMATCH_PROPS_OTHER:
        ANNOTMATCH_PROPS_OTHER.append(val)

ANNOTMATCH_PROPS_OTHER_SET = set([_.lower() for _ in ANNOTMATCH_PROPS_OTHER])
ANNOTMATCH_PROPS_OLD_SET = set([_.lower() for _ in OLD_ANNOTMATCH_PROPS])
# ANNOTMATCH_PROPS_STANDARD_SET = set([_.lower() for _ in ANNOTMATCH_PROPS_STANDARD])


def consolodate_annotmatch_tags(old_tags):
    # return case_tags
    remove_tags = [
        'hard',
        'needswork',
        'correctable',
        'uncorrectable',
        'interesting',
        'splitcase',
        'joincase',
        # 'orientation',
        'random',
        # 'badtail', 'badshoulder', 'splitcase', 'joincase', 'goodcoverage', 'interesting', 'hard'
    ]
    tags_dict = {
        # 'quality': 'Quality',
        # 'scoringissue': 'ScoringIssue',
        # 'orientation': 'Orientation',
        # 'orientation': 'MildViewpoint',
        'orientation': 'Viewpoint',
        # 'pose': 'SimilarPose',
        'pose': 'NonDistinct',
        # 'lighting': 'Lighting',
        # 'occlusion': 'Occlusion',
        # 'featurescale': 'FeatureScale',
        # 'edgematch': 'EdgeMatches',
        # 'featurescale': 'Pose',
        # 'featurescale': 'FeatureScale',
        'nondistinct': 'NonDistinct',
        'featurescale': 'NonDistinct',
        'edgematch': 'SimilarPose',
        'badtail': 'NonDistinct',
        'badshoulder': 'NonDistinct',
        # 'mildviewpoint': 'MildViewpoint',
        'mildviewpoint': 'Viewpoint',
        # 'toolargematches': 'CoarseFeatures',
        # 'badcoverage': 'LowCoverage',
        # 'shouldhavemore': 'LowCoverage',
        # 'viewpoint': 'Viewpoint',
    }

    def filter_tags(tags):
        return [t for t in tags if t.lower() not in remove_tags]

    def map_tags(tags):
        return [tags_dict.get(t.lower(), t) for t in tags]

    def cap_tags(tags):
        return [t[0].upper() + t[1:] for t in tags]

    filtered_tags = list(map(filter_tags, old_tags))
    mapped_tags = list(map(map_tags, filtered_tags))
    unique_tags = list(map(ut.unique_ordered, mapped_tags))
    new_tags = list(map(cap_tags, unique_tags))

    return new_tags


def rename_and_reduce_tags(ibs, annotmatch_rowids):
    """
    Script to update tags to newest values

    CommandLine:
        python -m wbia.tag_funcs --exec-rename_and_reduce_tags --db PZ_Master1

    Ignore:
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> #ibs = wbia.opendb(defaultdb='PZ_Master1')
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> annotmatch_rowids = filter_annotmatch_by_tags(ibs, min_num=1)
        >>> rename_and_reduce_tags(ibs, annotmatch_rowids)
    """
    tags_list_ = get_annotmatch_case_tags(ibs, annotmatch_rowids)

    def fix_tags(tags):
        return {six.text_type(t.lower()) for t in tags}

    tags_list = list(map(fix_tags, tags_list_))

    prop_mapping = {
        six.text_type(key.lower()): val for key, val in six.iteritems(PROP_MAPPING)
    }

    bad_tags = fix_tags(prop_mapping.keys())

    for rowid, tags in zip(annotmatch_rowids, tags_list):
        old_tags = tags.intersection(bad_tags)
        for tag in old_tags:
            ibs.set_annotmatch_prop(tag, [rowid], [False])
        new_tags = ut.dict_take(prop_mapping, old_tags)
        for tag in new_tags:
            if tag is not None:
                ibs.set_annotmatch_prop(tag, [rowid], [True])


def get_cate_categories():
    standard = ANNOTMATCH_PROPS_STANDARD
    other = ANNOTMATCH_PROPS_OTHER
    # case_list = standard + other
    return standard, other


def export_tagged_chips(ibs, aid_list, dpath='.'):
    """
    DEPRICATE

    CommandLine:
        python -m wbia.tag_funcs --exec-export_tagged_chips --tags Hard interesting needswork --db PZ_Master1
        python -m wbia.tag_funcs --exec-export_tagged_chips --logic=or --any_startswith quality occlusion --has_any lighting needswork interesting hard --db GZ_Master1 --dpath=/media/raid
        python -m wbia.tag_funcs --exec-export_tagged_chips --db GZ_Master1 --min_num=1  --dpath /media/raid

    Example:
        >>> # SCRIPT
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> kwargs = ut.argparse_dict(ut.get_kwdefaults2(filterflags_general_tags), type_hint=ut.ddict(list, logic=str))
        >>> ut.print_dict(kwargs, 'filter args')
        >>> aid_list = ibs.filter_annots_by_tags(**kwargs)
        >>> print('len(aid_list) = %r' % (len(aid_list),))
        >>> dpath = ut.get_argval('--dpath', default='')
        >>> all_tags = ut.flatten(ibs.get_annot_all_tags(aid_list))
        >>> filtered_tag_hist = ut.dict_hist(all_tags)
        >>> ut.print_dict(filtered_tag_hist, key_order_metric='val')
        >>> export_tagged_chips(ibs, aid_list, dpath)
    """
    visual_uuid_hashid = ibs.get_annot_hashid_visual_uuid(aid_list)
    zip_fpath = ut.unixjoin(
        dpath, 'exported_chips2_' + ibs.get_dbname() + visual_uuid_hashid + '.zip'
    )
    chip_fpath = ibs.get_annot_chip_fpath(aid_list)
    ut.archive_files(zip_fpath, chip_fpath, common_prefix=True)


@register_ibs_method
def filter_annots_by_tags(ibs, aid_list=None, **kwargs):
    """
    Filter / Find / Search for annotations with particular tags

    CommandLine:
        python -m wbia.tag_funcs --exec-filter_annots_by_tags --helpx
        python -m wbia.tag_funcs --exec-filter_annots_by_tags --db GZ_Master1
        python -m wbia.tag_funcs --exec-filter_annots_by_tags --db GZ_Master1 --min_num=1
        python -m wbia.tag_funcs --exec-filter_annots_by_tags --db GZ_Master1 --has_any=lighting --has_all=lighting:underexposed --show

    SeeAlso:
        python -m wbia.init.filter_annots --exec-filter_annots_general

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aid_list = ibs.get_valid_aids()
        >>> kwargs = ut.argparse_dict(ut.get_kwdefaults2(filterflags_general_tags), type_hint=ut.ddict(list, logic=str))
        >>> ut.print_dict(kwargs, 'filter args')
        >>> aid_list = ibs.filter_annots_by_tags(aid_list, **kwargs)
        >>> print('len(aid_list) = %r' % (len(aid_list),))
        >>> # print results
        >>> all_tags = ut.flatten(ibs.get_annot_all_tags(aid_list))
        >>> filtered_tag_hist = ut.dict_hist(all_tags)
        >>> ut.print_dict(filtered_tag_hist, key_order_metric='val')
        >>> print('len(aid_list) = %r' % (len(aid_list),))
        >>> print('sum(tags) = %r' % (sum(filtered_tag_hist.values()),))
        >>> ut.quit_if_noshow()
        >>> import wbia.viz.interact
        >>> wbia.viz.interact.interact_chip.interact_multichips(ibs, aid_list)
        >>> ut.show_if_requested()
    """
    if aid_list is None:
        aid_list = ibs.get_valid_aids()
    tags_list = ibs.get_annot_all_tags(aid_list)
    flags = filterflags_general_tags(tags_list, **kwargs)
    aid_list = ut.compress(aid_list, flags)
    return aid_list


@register_ibs_method
def filterflags_annot_tags(ibs, aid_list, **kwargs):
    """
    Filter / Find / Search for annotations with particular tags
    """
    tags_list = ibs.get_annot_all_tags(aid_list)
    flags = filterflags_general_tags(tags_list, **kwargs)
    return flags


@register_ibs_method
def get_aidpair_tags(ibs, aid1_list, aid2_list, directed=True):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        aid1_list (list):
        aid2_list (list):
        directed (bool): (default = True)

    Returns:
        list: tags_list

    CommandLine:
        python -m wbia.tag_funcs --exec-get_aidpair_tags --db PZ_Master1 --tags Hard interesting

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> has_any = ut.get_argval('--tags', type_=list, default=None)
        >>> min_num = ut.get_argval('--min_num', type_=int, default=1)
        >>> aid_pairs = filter_aidpairs_by_tags(ibs, has_any=has_any, min_num=1)
        >>> aid1_list = aid_pairs.T[0]
        >>> aid2_list = aid_pairs.T[1]
        >>> undirected_tags = get_aidpair_tags(ibs, aid1_list, aid2_list, directed=False)
        >>> tagged_pairs = list(zip(aid_pairs.tolist(), undirected_tags))
        >>> print(ut.repr2(tagged_pairs))
        >>> tag_dict = ut.groupby_tags(tagged_pairs, undirected_tags)
        >>> print(ut.repr2(tag_dict, nl=2))
        >>> print(ut.repr2(ut.map_dict_vals(len, tag_dict)))
    """
    aid_pairs = np.vstack([aid1_list, aid2_list]).T
    if directed:
        annotmatch_rowid = ibs.get_annotmatch_rowid_from_superkey(
            aid_pairs.T[0], aid_pairs.T[1]
        )
        tags_list = ibs.get_annotmatch_case_tags(annotmatch_rowid)
    else:
        annotmatch_rowid = ibs.get_annotmatch_rowid_from_undirected_superkey(
            aid_pairs.T[0], aid_pairs.T[1]
        )
        tags_list = ibs.get_annotmatch_case_tags(annotmatch_rowid)
        if False:
            expanded_aid_pairs = np.vstack([aid_pairs, aid_pairs[:, ::-1]])
            expanded_annotmatch_rowid = ibs.get_annotmatch_rowid_from_superkey(
                expanded_aid_pairs.T[0], expanded_aid_pairs.T[1]
            )
            expanded_edgeids = vt.get_undirected_edge_ids(expanded_aid_pairs)
            unique_edgeids, groupxs = vt.group_indices(expanded_edgeids)
            expanded_tags_list = ibs.get_annotmatch_case_tags(expanded_annotmatch_rowid)
            grouped_tags = vt.apply_grouping(
                np.array(expanded_tags_list, dtype=object), groupxs
            )
            undirected_tags = [list(set(ut.flatten(tags))) for tags in grouped_tags]
            edgeid2_tags = dict(zip(unique_edgeids, undirected_tags))
            input_edgeids = expanded_edgeids[: len(aid_pairs)]
            tags_list = ut.dict_take(edgeid2_tags, input_edgeids)
    return tags_list


@register_ibs_method
def filter_aidpairs_by_tags(
    ibs, has_any=None, has_all=None, min_num=None, max_num=None, am_rowids=None
):
    """
    list(zip(aid_pairs, undirected_tags))
    """
    # annotmatch_rowids = ibs.get_annotmatch_rowids_from_aid(aid_list)

    filtered_annotmatch_rowids = filter_annotmatch_by_tags(
        ibs,
        am_rowids,
        has_any=has_any,
        has_all=has_all,
        min_num=min_num,
        max_num=max_num,
    )
    aid1_list = np.array(ibs.get_annotmatch_aid1(filtered_annotmatch_rowids))
    aid2_list = np.array(ibs.get_annotmatch_aid2(filtered_annotmatch_rowids))
    aid_pairs = np.vstack([aid1_list, aid2_list]).T
    # Dont double count
    vt.get_undirected_edge_ids(aid_pairs)
    xs = vt.find_best_undirected_edge_indexes(aid_pairs)
    aid1_list = aid1_list.take(xs)
    aid2_list = aid2_list.take(xs)
    aid_pairs = np.vstack([aid1_list, aid2_list]).T
    return aid_pairs
    # directed_tags = get_aidpair_tags(ibs, aid_pairs.T[0], aid_pairs.T[1], directed=True)
    # valid_tags_list = ibs.get_annotmatch_case_tags(filtered_annotmatch_rowids)


def filter_annotmatch_by_tags(ibs, annotmatch_rowids=None, **kwargs):
    r"""
    ignores case

    Args:
        ibs (IBEISController):  wbia controller object
        flags (?):

    Returns:
        list

    CommandLine:
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db PZ_Master1 --min-num=1
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db PZ_Master1 --tags JoinCase
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db PZ_Master1 --tags SplitCase
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db PZ_Master1 --tags occlusion
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db PZ_Master1 --tags viewpoint
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db PZ_Master1 --tags SceneryMatch
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db PZ_Master1 --tags Photobomb

        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db GZ_Master1 --tags needswork

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> #ibs = wbia.opendb(defaultdb='testdb1')
        >>> ibs = wbia.opendb(defaultdb='PZ_Master1')
        >>> #tags = ['Photobomb', 'SceneryMatch']
        >>> has_any = ut.get_argval('--tags', type_=list, default=['SceneryMatch', 'Photobomb'])
        >>> min_num = ut.get_argval('--min_num', type_=int, default=1)
        >>> prop = has_any[0]
        >>> filtered_annotmatch_rowids = filter_annotmatch_by_tags(ibs, None, has_any=has_any, min_num=min_num)
        >>> aid1_list = np.array(ibs.get_annotmatch_aid1(filtered_annotmatch_rowids))
        >>> aid2_list = np.array(ibs.get_annotmatch_aid2(filtered_annotmatch_rowids))
        >>> aid_pairs = np.vstack([aid1_list, aid2_list]).T
        >>> # Dont double count
        >>> xs = vt.find_best_undirected_edge_indexes(aid_pairs)
        >>> aid1_list = aid1_list.take(xs)
        >>> aid2_list = aid2_list.take(xs)
        >>> valid_tags_list = ibs.get_annotmatch_case_tags(filtered_annotmatch_rowids)
        >>> print('valid_tags_list = %s' % (ut.repr2(valid_tags_list, nl=1),))
        >>> #
        >>> print('Aid pairs with has_any=%s' % (has_any,))
        >>> print('Aid pairs with min_num=%s' % (min_num,))
        >>> print('aid_pairs = ' + ut.repr2(list(zip(aid1_list, aid2_list))))
        >>> # Show timedelta info
        >>> ut.quit_if_noshow()
        >>> timedelta_list = ibs.get_annot_pair_timedelta(aid1_list, aid2_list)
        >>> import wbia.plottool as pt
        >>> pt.draw_timedelta_pie(timedelta_list, label='timestamp of tags=%r' % (has_any,))
        >>> ut.show_if_requested()
    """
    if annotmatch_rowids is None:
        annotmatch_rowids = ibs._get_all_annotmatch_rowids()

    tags_list = ibs.get_annotmatch_case_tags(annotmatch_rowids)

    flags = filterflags_general_tags(tags_list, **kwargs)
    filtered_annotmatch_rowids = ut.compress(annotmatch_rowids, flags)
    return filtered_annotmatch_rowids


# TODO: ut.filterflags_general_tags


def filterflags_general_tags(
    tags_list,
    has_any=None,
    has_all=None,
    has_none=None,
    min_num=None,
    max_num=None,
    any_startswith=None,
    any_endswith=None,
    any_match=None,
    none_match=None,
    logic='and',
):
    r"""
    maybe integrate into utool? Seems pretty general

    Args:
        tags_list (list):
        has_any (None): (default = None)
        has_all (None): (default = None)
        min_num (None): (default = None)
        max_num (None): (default = None)

    CommandLine:
        python -m wbia.tag_funcs --exec-filterflags_general_tags
        python -m wbia.tag_funcs --exec-filterflags_general_tags:0  --helpx
        python -m wbia.tag_funcs --exec-filterflags_general_tags:0
        python -m wbia.tag_funcs --exec-filterflags_general_tags:0  --none_match n
        python -m wbia.tag_funcs --exec-filterflags_general_tags:0  --has_none=n,o
        python -m wbia.tag_funcs --exec-filterflags_general_tags:1
        python -m wbia.tag_funcs --exec-filterflags_general_tags:2

    Example0:
        >>> # DISABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> tags_list = [['v'], [], ['P'], ['P', 'o'], ['n', 'o',], [], ['n', 'N'], ['e', 'i', 'p', 'b', 'n'], ['q', 'v'], ['n'], ['n'], ['N']]
        >>> kwargs = ut.argparse_dict(ut.get_kwdefaults2(filterflags_general_tags), type_hint=list)
        >>> print('kwargs = %r' % (kwargs,))
        >>> flags = filterflags_general_tags(tags_list, **kwargs)
        >>> print(flags)
        >>> result = ut.compress(tags_list, flags)
        >>> print('result = %r' % (result,))

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> tags_list = [['v'], [], ['P'], ['P'], ['n', 'o',], [], ['n', 'N'], ['e', 'i', 'p', 'b', 'n'], ['n'], ['n'], ['N']]
        >>> has_all = 'n'
        >>> min_num = 1
        >>> flags = filterflags_general_tags(tags_list, has_all=has_all, min_num=min_num)
        >>> result = ut.compress(tags_list, flags)
        >>> print('result = %r' % (result,))

    Example2:
        >>> # ENABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> tags_list = [['vn'], ['vn', 'no'], ['P'], ['P'], ['n', 'o',], [], ['n', 'N'], ['e', 'i', 'p', 'b', 'n'], ['n'], ['n', 'nP'], ['NP']]
        >>> kwargs = {
        >>>     'any_endswith': 'n',
        >>>     'any_match': None,
        >>>     'any_startswith': 'n',
        >>>     'has_all': None,
        >>>     'has_any': None,
        >>>     'has_none': None,
        >>>     'max_num': 3,
        >>>     'min_num': 1,
        >>>     'none_match': ['P'],
        >>> }
        >>> flags = filterflags_general_tags(tags_list, **kwargs)
        >>> filtered = ut.compress(tags_list, flags)
        >>> result = ('result = %s' % (ut.repr2(filtered),))
        result = [['vn', 'no'], ['n', 'o'], ['n', 'N'], ['n'], ['n', 'nP']]
    """
    import re
    import operator

    def fix_tags(tags):
        return {six.text_type(t.lower()) for t in tags}

    if logic is None:
        logic = 'and'

    logic_func = {'and': np.logical_and, 'or': np.logical_or}[logic]

    default_func = {'and': np.ones, 'or': np.zeros}[logic]

    tags_list_ = [fix_tags(tags_) for tags_ in tags_list]
    flags = default_func(len(tags_list_), dtype=np.bool)

    if min_num is not None:
        flags_ = [len(tags_) >= min_num for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    if max_num is not None:
        flags_ = [len(tags_) <= max_num for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    if has_any is not None:
        has_any = fix_tags(set(ut.ensure_iterable(has_any)))
        flags_ = [len(has_any.intersection(tags_)) > 0 for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    if has_none is not None:
        has_none = fix_tags(set(ut.ensure_iterable(has_none)))
        flags_ = [len(has_none.intersection(tags_)) == 0 for tags_ in tags_list_]
        logic_func(flags, flags_, out=flags)

    if has_all is not None:
        has_all = fix_tags(set(ut.ensure_iterable(has_all)))
        flags_ = [
            len(has_all.intersection(tags_)) == len(has_all) for tags_ in tags_list_
        ]
        logic_func(flags, flags_, out=flags)

    def check_item(tags_, fields, op, compare):
        t_flags = [any([compare(t, f) for f in fields]) for t in tags_]
        num_passed = sum(t_flags)
        flag = op(num_passed, 0)
        return flag

    def flag_tags(tags_list, fields, op, compare):
        flags = [check_item(tags_, fields, op, compare) for tags_ in tags_list_]
        return flags

    def execute_filter(flags, tags_list, fields, op, compare):
        if fields is not None:
            fields = ut.ensure_iterable(fields)
            flags_ = flag_tags(tags_list, fields, op, compare)
            logic_func(flags, flags_, out=flags)
        return flags

    flags = execute_filter(
        flags, tags_list, any_startswith, operator.gt, six.text_type.startswith
    )

    flags = execute_filter(
        flags, tags_list, any_endswith, operator.gt, six.text_type.endswith
    )

    flags = execute_filter(
        flags, tags_list, any_match, operator.gt, lambda t, f: re.match(f, t)
    )

    flags = execute_filter(
        flags, tags_list, none_match, operator.eq, lambda t, f: re.match(f, t)
    )
    return flags


@register_ibs_method
@profile
def get_annotmatch_case_tags(ibs, annotmatch_rowids):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        annotmatch_rowids (?):

    Returns:
        list: filtered_aid_list

    CommandLine:
        python -m wbia.tag_funcs --exec-get_annotmatch_case_tags

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='PZ_Master1')
        >>> annotmatch_rowids = ibs._get_all_annotmatch_rowids()
        >>> tags_list = get_annotmatch_case_tags(ibs, annotmatch_rowids)
        >>> result = ('tags_list = %s' % (str(tags_list),))
        >>> print(result)
        tags_list = [[u'occlusion', u'pose', 'Hard', 'NonDistinct'], [], ['Hard']]
    """
    standard, other = get_cate_categories()
    annotmatch_tag_texts_list = ibs.get_annotmatch_tag_text(annotmatch_rowids)
    tags_list = [
        [] if note is None else _parse_tags(note) for note in annotmatch_tag_texts_list
    ]
    # NEW = False
    # if NEW:
    #    # hack for faster tag parsing
    #    from wbia.control import _autogen_annotmatch_funcs as _aaf
    #    import itertools
    #    colnames = (_aaf.ANNOTMATCH_IS_HARD, _aaf.ANNOTMATCH_IS_SCENERYMATCH,
    #                _aaf.ANNOTMATCH_IS_PHOTOBOMB, _aaf.ANNOTMATCH_IS_NONDISTINCT)
    #    id_iter = annotmatch_rowids
    #    annotmatch_is_col = ibs.db.get(
    #        ibs.const.ANNOTMATCH_TABLE, colnames, id_iter, id_colname='rowid',
    #        eager=True, nInput=None, unpack_scalars=True)
    #    annotmatch_is_col = [col if col is not None else [None] * len(colnames)
    #                         for col in annotmatch_is_col]
    #    standardtags = [x[len('annotmatch_is_'):] for x in colnames]
    #    standard_tags_list = ut.list_zipcompress(itertools.repeat(standardtags), annotmatch_is_col)
    #    tags_list = [tags1 + tags2 for tags1, tags2 in zip(tags_list, standard_tags_list)]
    # else:
    #    for case in standard:
    #        flag_list = ibs.get_annotmatch_prop(case, annotmatch_rowids)
    #        for tags in ut.compress(tags_list, flag_list):
    #            tags.append(case)
    tags_list = [[six.text_type(t) for t in tags] for tags in tags_list]
    # if ut.get_argval('--consol') or True:
    #    tags_list = consolodate_annotmatch_tags(tags_list)
    return tags_list


@profile
def get_annotmatch_standard_prop(ibs, prop, annotmatch_rowids):
    getter = getattr(ibs, 'get_annotmatch_is_' + prop.lower())
    flag_list = getter(annotmatch_rowids)
    return flag_list


@register_ibs_method
@profile
def get_annotmatch_prop(ibs, prop, annotmatch_rowids):
    r"""
    hacky getter for dynamic properties of annotmatches using notes table

    Args:
        prop (str):
        annotmatch_rowids (?):

    Returns:
        list: filtered_aid_list

    CommandLine:
        python -m wbia.tag_funcs --exec-get_annotmatch_prop

    Example:
        >>> # DISABLE_DOCTEST
        >>> # Test setting and getting standard keys
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> prop = 'hard'
        >>> annotmatch_rowids = ibs._get_all_annotmatch_rowids()
        >>> flag_list = get_annotmatch_prop(ibs, prop, annotmatch_rowids)
        >>> flag_list = ('filtered_aid_list = %s' % (str(flag_list),))
        >>> subset_rowids = annotmatch_rowids[::2]
        >>> set_annotmatch_prop(ibs, prop, subset_rowids, [True] * len(subset_rowids))
        >>> flag_list2 = get_annotmatch_prop(ibs, prop, annotmatch_rowids)
        >>> print('flag_list2 = %r' % (flag_list2,))

    Example:
        >>> # DISABLE_DOCTEST
        >>> # Test setting and getting non-standard keys
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> prop = 'occlusion'
        >>> annotmatch_rowids = ibs._get_all_annotmatch_rowids()
        >>> flag_list = get_annotmatch_prop(ibs, prop, annotmatch_rowids)
        >>> flag_list = ('filtered_aid_list = %s' % (str(flag_list),))
        >>> subset_rowids = annotmatch_rowids[1::2]
        >>> subset_rowids1 = annotmatch_rowids[::2]
        >>> set_annotmatch_prop(ibs, prop, subset_rowids1, [True] * len(subset_rowids))
        >>> set_annotmatch_prop(ibs, 'pose', subset_rowids1, [True] * len(subset_rowids))
        >>> flag_list2 = get_annotmatch_prop(ibs, prop, annotmatch_rowids)
        >>> print('flag_list2 = %r' % (flag_list2,))
    """
    # if prop.lower() in ANNOTMATCH_PROPS_STANDARD_SET:
    #    return ibs.get_annotmatch_standard_prop(prop, annotmatch_rowids)
    for prop_ in ut.ensure_iterable(prop):
        flag1 = prop_.lower() not in ANNOTMATCH_PROPS_OTHER_SET
        flag2 = prop_.lower() not in ANNOTMATCH_PROPS_OLD_SET
        if flag1 and flag2:
            raise NotImplementedError('Unknown prop_=%r' % (prop_,))
    return get_annotmatch_other_prop(ibs, prop, annotmatch_rowids)


@register_ibs_method
def set_annotmatch_prop(ibs, prop, annotmatch_rowids, flags):
    """
    hacky setter for dynamic properties of annotmatches using notes table
    """
    logger.info(
        '[ibs] set_annotmatch_prop prop=%s for %d pairs' % (prop, len(annotmatch_rowids))
    )
    # if prop.lower() in ANNOTMATCH_PROPS_STANDARD_SET:
    #    setter = getattr(ibs, 'set_annotmatch_is_' + prop.lower())
    #    return setter(annotmatch_rowids, flags)
    if (
        prop.lower() in ANNOTMATCH_PROPS_OTHER_SET
        or prop.lower() in ANNOTMATCH_PROPS_OLD_SET
    ):
        return set_annotmatch_other_prop(ibs, prop, annotmatch_rowids, flags)
    else:
        raise NotImplementedError(
            'Unknown prop=%r not in %r' % (prop, ANNOTMATCH_PROPS_OTHER_SET)
        )


def _parse_tags(note):
    """ convert a note into tags """
    return [tag.strip() for tag in note.split(';') if len(tag) > 0]


def _remove_tag(tags, prop):
    """ convert a note into tags """
    try:
        tags.remove(prop)
    except ValueError:
        pass
    return tags


@profile
def get_annotmatch_other_prop(ibs, prop, annotmatch_rowids):
    annotmatch_tag_texts_list = ibs.get_annotmatch_tag_text(annotmatch_rowids)
    flag_list = get_textformat_tag_flags(prop, annotmatch_tag_texts_list)
    return flag_list


def set_annotmatch_other_prop(ibs, prop, annotmatch_rowids, flags):
    """
    sets nonstandard properties using the notes column
    """
    annotmatch_tag_texts_list = ibs.get_annotmatch_tag_text(annotmatch_rowids)
    new_notes_list = set_textformat_tag_flags(prop, annotmatch_tag_texts_list, flags)
    ibs.set_annotmatch_tag_text(annotmatch_rowids, new_notes_list)


@profile
def get_textformat_tag_flags(prop, text_list):
    """ general text tag getter hack """
    tags_list = [None if note is None else _parse_tags(note) for note in text_list]
    if ut.isiterable(prop):
        props_ = [p.lower() for p in prop]
        flags_list = [
            [None if tags is None else int(prop_ in tags) for tags in tags_list]
            for prop_ in props_
        ]
        return flags_list
    else:
        prop = prop.lower()
        flag_list = [None if tags is None else int(prop in tags) for tags in tags_list]
        return flag_list


def set_textformat_tag_flags(prop, text_list, flags):
    """ general text tag setter hack """
    prop = prop.lower()
    ensured_text = ['' if note is None else note for note in text_list]
    tags_list = [_parse_tags(note) for note in ensured_text]
    # Remove from all
    new_tags_list = [_remove_tag(tags, prop) for tags in tags_list]
    # then add to specified ones
    for tags, flag in zip(new_tags_list, flags):
        if flag:
            tags.append(prop)
    new_text_list = [';'.join(tags) for tags in new_tags_list]
    return new_text_list


ANNOT_TAGS = [
    'occlusion',
    'lighting',
    'quality',
    'pose',
    'error',
    'interesting',
    'error:viewpoint',
    'error:quality',
    'occlusion:large',
    'occlusion:medium',
    'occlusion:small',
    'lighting:shadowed',
    'lighting:overexposed',
    'lighting:underexposed',
    'quality:washedout',
    'quality:blury',
    'pose:novel',
    'pose:common',
    'error:bbox',
    'error:mask',
    'error:other',
]


def get_available_annot_tags():
    return ANNOT_TAGS


def get_annot_prop(ibs, prop, aid_list):
    """
    Annot tags
    """
    text_list = ibs.get_annot_tag_text(aid_list)
    flag_list = get_textformat_tag_flags(prop, text_list)
    return flag_list


@register_ibs_method
def set_annot_prop(ibs, prop, aid_list, flags):
    """
    sets nonstandard properties using the notes column
    """
    text_list = ibs.get_annot_tag_text(aid_list)
    new_text_list = set_textformat_tag_flags(prop, text_list, flags)
    ibs.set_annot_tag_text(aid_list, new_text_list)


@register_ibs_method
def append_annot_case_tags(ibs, aid_list, tag_list):
    """
    Generally appends tags to annotations. Careful not to introduce too many
    random tags. Maybe we should just let that happen and introduce tag-aliases

    Note: this is more of a set add rather than a list append

    TODO: remove
    """
    # Ensure each item is a list
    # tags_list = [tag if isinstance(tag, list) else [tag] for tag in tag_list]
    if isinstance(tag_list, six.string_types):
        # Apply single tag to everybody
        tag_list = [tag_list] * len(aid_list)
    tags_list = [ut.ensure_iterable(tag) for tag in tag_list]
    text_list = ibs.get_annot_tag_text(aid_list)
    orig_tags_list = [[] if note is None else _parse_tags(note) for note in text_list]
    new_tags_list = [ut.unique(t1 + t2) for t1, t2 in zip(tags_list, orig_tags_list)]
    ibs.set_annot_case_tags(aid_list, new_tags_list)


@register_ibs_method
def set_annot_case_tags(ibs, aid_list, new_tags_list):
    """
    Completely overwrite case tags
    """
    for tag in new_tags_list:
        assert isinstance(tag, list), 'each set of tags must be a list of strs'
    new_text_list = [';'.join(tags) for tags in new_tags_list]
    ibs.set_annot_tag_text(aid_list, new_text_list)


@register_ibs_method
def remove_annot_case_tags(ibs, aid_list, tag_list):
    if isinstance(tag_list, six.string_types):
        # Apply single tag to everybody
        tag_list = [tag_list] * len(aid_list)
    tags_list = [ut.ensure_iterable(tag) for tag in tag_list]
    text_list = ibs.get_annot_tag_text(aid_list)
    orig_tags_list = [[] if note is None else _parse_tags(note) for note in text_list]
    new_tags_list = [ut.setdiff(t2, t1) for t1, t2 in zip(tags_list, orig_tags_list)]
    new_text_list = [';'.join(tags) for tags in new_tags_list]
    ibs.set_annot_tag_text(aid_list, new_text_list)


@register_ibs_method
def overwrite_annot_case_tags(ibs, aid_list, tag_list):
    """
    Completely replaces annotation tags.
    BE VERY CAREFUL WITH THIS FUNCTION
    """
    assert all([ut.isiterable(tag) for tag in tag_list])
    # text_list = ibs.get_annot_tag_text(aid_list)
    # orig_tags_list = [[] if note is None else _parse_tags(note) for note in text_list]
    new_tags_list = tag_list
    new_text_list = [';'.join(tags) for tags in new_tags_list]
    ibs.set_annot_tag_text(aid_list, new_text_list)


@register_ibs_method
def remove_all_annot_case_tags(ibs, aid_list):
    ibs.set_annot_tag_text(aid_list, [''] * len(aid_list))


@register_ibs_method
def get_annot_case_tags(ibs, aid_list):
    r"""
    returns list of tags. Use instead of get_annot_tag_text
    TODO:
        rename to get_annot_unary_tags

    Args:
        ibs (IBEISController):  wbia controller object
        aid_list (list):  list of annotation rowids

    Returns:
        list: tags_list

    CommandLine:
        python -m wbia.tag_funcs --exec-get_annot_case_tags

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> from wbia.tag_funcs import _parse_tags # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aid_list = ibs.get_valid_aids()
        >>> tags_list = get_annot_case_tags(ibs, aid_list)
        >>> result = ('tags_list = %s' % (str(tags_list),))
        >>> print(result)

    Ignore:
        # FIXME incorrporate old tag notes
        aid_list = ibs.get_valid_aids()
        notes_list = ibs.get_annot_notes(aid_list)
        flags = [len(notes) > 0 for notes in notes_list]
        aid_list = ut.compress(aid_list, flags)
        notes_list = ut.compress(notes_list, flags)

        import re
        notes_list = [note.replace('rfdetect', '') for note in notes_list]
        notes_list = [note.replace('<COMMA>', ';') for note in notes_list]
        notes_list = [note.replace('jpg', '') for note in notes_list]
        notes_list = [note.replace('<HARDCASE>', '') for note in notes_list]
        notes_list = [note.strip() for note in notes_list]
        notes_list = [re.sub(';;*', ';', note) for note in notes_list]
        notes_list = [note.strip(';') for note in notes_list]
        notes_list = [note.strip(':') for note in notes_list]
        notes_list = [note.strip() for note in notes_list]

        flags = [len(notes) < 70 and len(notes) > 0 for notes in notes_list]
        aid_list = ut.compress(aid_list, flags)
        notes_list = ut.compress(notes_list, flags)

        flags = ['M;' not in notes and 'F;' not in notes and 'H1' not in notes for notes in notes_list]
        flags = [ 'M;' not in notes and 'F;' not in notes and 'H1' not in notes for notes in notes_list]
        aid_list = ut.compress(aid_list, flags)
        notes_list = ut.compress(notes_list, flags)

        flags = ['aliases' not in notes for notes in notes_list]
        aid_list = ut.compress(aid_list, flags)
        notes_list = ut.compress(notes_list, flags)

        #flags = [not re.match(';\d*;', note) for note  in notes_list]
        flags = [not re.match(r'\d\d*', note) for note  in notes_list]
        aid_list = ut.compress(aid_list, flags)
        notes_list = ut.compress(notes_list, flags)

        flags = [not notes.startswith('Foal;') for notes in notes_list]
        aid_list = ut.compress(aid_list, flags)
        notes_list = ut.compress(notes_list, flags)

        old_tags_list = [_parse_tags(note) for note in notes_list]

        old_tags = list(set(ut.flatten(old_tags_list)))
        old_tags = sorted([tag for tag in old_tags if not re.match(r'\d\d*', tag)])

        old_to_new = {
            'gash':  None,
            'pose':  'novelpose',
            'vocalizing': 'novelpose'
            'occlusion':  'occlusion',
        }

    Ignore:
        python -m wbia.tag_funcs --exec-filter_annotmatch_by_tags --show --db PZ_Master1 --tags viewpoint

    """
    text_list = ibs.get_annot_tag_text(aid_list)
    tags_list = [[] if note is None else _parse_tags(note) for note in text_list]
    return tags_list


@register_ibs_method
@profile
def get_annot_annotmatch_tags(ibs, aid_list):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        aid_list (list):  list of annotation rowids

    Returns:
        list: annotmatch_tags_list

    CommandLine:
        python -m wbia.tag_funcs --exec-get_annot_annotmatch_tags --db GZ_Master1

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aid_list = ibs.get_valid_aids()
        >>> all_tags = ut.flatten(get_annot_annotmatch_tags(ibs, aid_list))
        >>> tag_hist = ut.dict_hist(all_tags)
        >>> ut.print_dict(tag_hist)
    """
    annotmatch_rowids = ibs.get_annotmatch_rowids_from_aid(aid_list)
    unflat_tags_list = ibs.unflat_map(ibs.get_annotmatch_case_tags, annotmatch_rowids)
    annotmatch_tags_list = [
        list(set(ut.flatten(_unflat_tags))) for _unflat_tags in unflat_tags_list
    ]
    return annotmatch_tags_list


@register_ibs_method
@profile
def get_annot_all_tags(ibs, aid_list=None):
    """
    CommandLine:
        python -m wbia.tag_funcs --exec-get_annot_all_tags --db GZ_Master1

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.tag_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aid_list = ibs.get_valid_aids()
        >>> all_tags = ut.flatten(ibs.get_annot_all_tags(aid_list))
        >>> tag_hist = ut.dict_hist(all_tags)
        >>> ut.print_dict(tag_hist)
    """
    if aid_list is None:
        aid_list = ibs.get_valid_aids()

    annotmatch_tags_list = ibs.get_annot_annotmatch_tags(aid_list)
    annot_tags_list = ibs.get_annot_case_tags(aid_list)
    both_tags_list = list(
        map(
            ut.unique_ordered,
            map(ut.flatten, zip(annot_tags_list, annotmatch_tags_list)),
        )
    )
    return both_tags_list



# -*- coding: utf-8 -*-
"""Operations for dumping a database to file"""
__all__ = (
    'dump',
    'dumps',
)


def _iter_dump(db_connection, schema_only=False):
    """\
    Generator function that serializes the database associated with
    ``db_connection`` as formatted ``str``.

    Args:
        db_connection (sqlite3.Connection): database instance to dump
        schema_only (bool): flag to only dump the schema

    Returns:
        str: serialized string of database

    """
    for line in db_connection.iterdump():
        # ??? (23-Jul-12020) This seems a bit dicey, good idea?
        #     This may only work in sqlite environment.
        if schema_only and line.startswith('INSERT'):
            # FIXME (23-Jul-12020) hardcoded table name, use METADATA_TABLE
            #       when circular import is resolved.
            # if metadata is requested then allow those inserts
            if 'INSERT INTO "metadata' not in line:
                continue
        yield f'{line}\n'


def dumps(*args, **kwargs):
    """\
    Serialize the database associated with ``db_connection`` as formatted ``str``.

    Args:
        db_connection (sqlite3.Connection): database instance to dump
        schema_only (bool): flag to only dump the schema

    Returns:
        str: serialized string of database

    """

    return ''.join([s for s in _iter_dump(*args, **kwargs)])


def dump(db_connection, fp, **kwargs):
    """\
    Serialize the database associated with ``db_connection`` as a formatted
    stream to ``fp`` (a ``.write()``-supporting file-like object).

    This is a side-effect only function that writes to ``fp``.

    Args:
        db_connection (sqlite3.Connection): database instance to dump
        fp (file-like object): output stream
        schema_only (bool): flag to only dump the schema

    Returns:
        None

    """
    for line in _iter_dump(db_connection, **kwargs):
        fp.write(line)



# -*- coding: utf-8 -*-
import logging
import uuid
import six  # NOQA
from six.moves import range
from wbia import constants as const
from wbia.control.accessor_decors import (
    adder,
    getter_1to1,
    getter_1toM,
)
import utool as ut

# from wbia.other import ibsfuncs
from wbia.control.controller_inject import make_ibs_register_decorator

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


# ADDERS::IMAGE->IMAGESET

#
# GETTERS::GLR


@register_ibs_method
@getter_1to1
def get_glr_confidence(ibs, glrid_list):
    """
    Returns:
        list_ (list):  confidence in an image relationship"""
    glr_confidence_list = ibs.db.get(
        const.GL_RELATION_TABLE, ('glr_confidence',), glrid_list
    )
    return glr_confidence_list


@register_ibs_method
@getter_1to1
def get_glr_lblimage_rowids(ibs, glrid_list):
    """ get the lblimage_rowid belonging to each relationship """
    lblimage_rowids_list = ibs.db.get(
        const.GL_RELATION_TABLE, ('lblimage_rowid',), glrid_list
    )
    return lblimage_rowids_list


@register_ibs_method
@getter_1to1
def get_glr_image_rowids(ibs, glrid_list):
    """ get the image_rowid belonging to each relationship """
    image_rowids_list = ibs.db.get(const.GL_RELATION_TABLE, ('image_rowid',), glrid_list)
    return image_rowids_list


# ADDERS::LBLIMAGE


@register_ibs_method
@adder
def add_lblimages(
    ibs, lbltype_rowid_list, value_list, note_list=None, lblimage_uuid_list=None
):
    """Adds new lblimages (labels of imageations)
    creates a new uuid for any new pair(type, value)
    #TODO: reverse order of rowid_list value_list in input
    """
    if note_list is None:
        note_list = [''] * len(value_list)
    # Get random uuids
    if lblimage_uuid_list is None:
        lblimage_uuid_list = [uuid.uuid4() for _ in range(len(value_list))]
    colnames = ['lblimage_uuid', 'lbltype_rowid', 'lblimage_value', 'lblimage_note']
    params_iter = list(zip(lblimage_uuid_list, lbltype_rowid_list, value_list, note_list))
    get_rowid_from_superkey = ibs.get_lblimage_rowid_from_superkey
    superkey_paramx = (1, 2)
    lblimage_rowid_list = ibs.db.add_cleanly(
        const.LBLIMAGE_TABLE,
        colnames,
        params_iter,
        get_rowid_from_superkey,
        superkey_paramx,
    )
    return lblimage_rowid_list


#
# GETTERS::LBLIMAGE_TABLE


@register_ibs_method
@getter_1to1
def get_lblimage_rowid_from_superkey(ibs, lbltype_rowid_list, value_list):
    """
    Returns:
        list_ (list):  lblimage_rowid_list from the superkey (lbltype, value)
    """
    colnames = ('lblimage_rowid',)
    params_iter = zip(lbltype_rowid_list, value_list)
    where_clause = 'lbltype_rowid=? AND lblimage_value=?'
    lblimage_rowid_list = ibs.db.get_where(
        const.LBLIMAGE_TABLE, colnames, params_iter, where_clause
    )
    return lblimage_rowid_list


@register_ibs_method
@getter_1to1
def get_lblimage_rowid_from_uuid(ibs, lblimage_uuid_list):
    """
    Returns:
        list_ (list):  lblimage_rowid_list from the superkey (lbltype, value)
    """
    colnames = ('lblimage_rowid',)
    params_iter = lblimage_uuid_list
    id_colname = 'lblimage_uuid'
    lblimage_rowid_list = ibs.db.get(
        const.LBLIMAGE_TABLE, colnames, params_iter, id_colname=id_colname
    )
    return lblimage_rowid_list


@register_ibs_method
@getter_1to1
def get_lblimage_uuids(ibs, lblimage_rowid_list):
    lblimageuuid_list = ibs.db.get(
        const.LBLIMAGE_TABLE, ('lblimage_uuid',), lblimage_rowid_list
    )
    return lblimageuuid_list


@register_ibs_method
@getter_1to1
def get_lblimage_lbltypes_rowids(ibs, lblimage_rowid_list):
    lbltype_rowid_list = ibs.db.get(
        const.LBLIMAGE_TABLE, ('lbltype_rowid',), lblimage_rowid_list
    )
    return lbltype_rowid_list


@register_ibs_method
@getter_1to1
def get_lblimage_notes(ibs, lblimage_rowid_list):
    lblimagenotes_list = ibs.db.get(
        const.LBLIMAGE_TABLE, ('lblimage_note',), lblimage_rowid_list
    )
    return lblimagenotes_list


@register_ibs_method
@getter_1to1
def get_lblimage_values(ibs, lblimage_rowid_list, _lbltype=None):
    """
    Returns:
        list_ (list): text lblimages"""
    # TODO: Remove keyword argument
    # ibsfuncs.assert_lblimage_rowids_are_type(ibs, lblimage_rowid_list,  ibs.lbltype_ids[_lbltype])
    lblimage_value_list = ibs.db.get(
        const.LBLIMAGE_TABLE, ('lblimage_value',), lblimage_rowid_list
    )
    return lblimage_value_list


@register_ibs_method
def get_lblimage_gids(ibs, lblimage_rowid_list):
    # verbose = len(lblimage_rowid_list) > 20
    # TODO: Optimize IF POSSIBLE
    # FIXME: SLOW
    # if verbose:
    #    logger.info(ut.get_caller_name(N=list(range(0, 20))))
    where_clause = 'lblimage_rowid=?'
    params_iter = [(lblimage_rowid,) for lblimage_rowid in lblimage_rowid_list]
    gids_list = ibs.db.get_where(
        const.GL_RELATION_TABLE,
        ('image_rowid',),
        params_iter,
        where_clause,
        unpack_scalars=False,
    )
    return gids_list


# ADDERS::GLR


@register_ibs_method
@adder
def add_image_relationship_one(
    ibs, gid_list, lblimage_rowid_list, glr_confidence_list=None
):
    """Adds a relationship between images and lblimages
    (imageations and labels of imageations)"""
    if glr_confidence_list is None:
        glr_confidence_list = [0.0] * len(gid_list)
    colnames = (
        'image_rowid',
        'lblimage_rowid',
        'glr_confidence',
    )
    params_iter = list(zip(gid_list, lblimage_rowid_list, glr_confidence_list))
    get_rowid_from_superkey = ibs.get_glrid_from_superkey
    superkey_paramx = (0, 1)  # TODO HAVE SQL GIVE YOU THESE NUMBERS
    glrid_list = ibs.db.add_cleanly(
        const.GL_RELATION_TABLE,
        colnames,
        params_iter,
        get_rowid_from_superkey,
        superkey_paramx,
    )
    return glrid_list


@register_ibs_method
@getter_1to1
def get_glrid_from_superkey(ibs, gid_list, lblimage_rowid_list):
    """
    Args:
        gid_list (list): list of image row-ids
        lblimage_rowid_list (list): list of lblimage row-ids
    Returns:
        glrid_list (list): image-label relationship id list
    """
    colnames = ('image_rowid',)
    params_iter = zip(gid_list, lblimage_rowid_list)
    where_clause = 'image_rowid=? AND lblimage_rowid=?'
    glrid_list = ibs.db.get_where(
        const.GL_RELATION_TABLE, colnames, params_iter, where_clause
    )
    return glrid_list


@register_ibs_method
@getter_1toM
def get_image_glrids(ibs, gid_list):
    """FIXME: __name__
    Get all the relationship ids belonging to the input images
    if lblimage lbltype is specified the relationship ids are filtered to
    be only of a specific lbltype/category/type
    """
    params_iter = ((gid,) for gid in gid_list)
    where_clause = 'image_rowid=?'
    glrids_list = ibs.db.get_where(
        const.GL_RELATION_TABLE,
        ('glr_rowid',),
        params_iter,
        where_clause=where_clause,
        unpack_scalars=False,
    )
    return glrids_list



# -*- coding: utf-8 -*-
# flake8: noqa
import logging
import utool

(print, rrr, profile) = utool.inject2(__name__, '[web]')
logger = logging.getLogger('wbia')

from wbia.web import apis_detect
from wbia.web import apis_engine
from wbia.web import apis_json
from wbia.web import apis_sync
from wbia.web import apis_query
from wbia.web import apis
from wbia.web import app
from wbia.web import appfuncs
from wbia.web import routes_ajax
from wbia.web import routes_demo
from wbia.web import routes_csv
from wbia.web import routes_experiments
from wbia.web import routes_submit
from wbia.web import routes


from wbia.control import controller_inject

if controller_inject.MICROSOFT_API_ENABLED:
    from wbia.web import apis_microsoft



# -*- coding: utf-8 -*-
import logging
from wbia import constants as const
from wbia.control.accessor_decors import adder, getter_1to1, ider
import utool as ut
from wbia.control.controller_inject import make_ibs_register_decorator

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


@register_ibs_method
@ider
def _get_all_known_lblannot_rowids(ibs, _lbltype):
    """
    Returns:
        list_ (list): all nids of known animals
        (does not include unknown names)"""
    all_known_lblannot_rowids = ibs.db.get_all_rowids_where(
        const.LBLANNOT_TABLE, 'lbltype_rowid=?', (ibs.lbltype_ids[_lbltype],)
    )
    return all_known_lblannot_rowids


@register_ibs_method
@adder
def add_lbltype(ibs, text_list, default_list):
    """Adds a label type and its default value
    Should only be called at the begining of the program.
    """
    params_iter = zip(text_list, default_list)
    colnames = (
        'lbltype_text',
        'lbltype_default',
    )
    get_rowid_from_superkey = ibs.get_lbltype_rowid_from_text
    lbltype_rowid_list = ibs.db.add_cleanly(
        const.LBLTYPE_TABLE, colnames, params_iter, get_rowid_from_superkey
    )
    return lbltype_rowid_list


#
# GETTERS::LBLTYPE


@register_ibs_method
@getter_1to1
def get_lbltype_rowid_from_text(ibs, text_list):
    """
    Returns:
        lbltype_rowid (list): lbltype_rowid where the lbltype_text is given
    """
    # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
    # FIXME: Use unique SUPERKEYS instead of specifying id_colname
    lbltype_rowid = ibs.db.get(
        const.LBLTYPE_TABLE, ('lbltype_rowid',), text_list, id_colname='lbltype_text'
    )
    return lbltype_rowid


@register_ibs_method
@getter_1to1
def get_lbltype_default(ibs, lbltype_rowid_list):
    lbltype_default_list = ibs.db.get(
        const.LBLTYPE_TABLE, ('lbltype_default',), lbltype_rowid_list
    )
    return lbltype_default_list


@register_ibs_method
@getter_1to1
def get_lbltype_text(ibs, lbltype_rowid_list):
    lbltype_text_list = ibs.db.get(
        const.LBLTYPE_TABLE, ('lbltype_text',), lbltype_rowid_list
    )
    return lbltype_text_list
